import numpy as np
import pandas as pd


def read_goto(filename: str) -> pd.DataFrame:
    """Import and parse route waypoints in a goto file.

    Parameters
    ----------
    filename : str
        The file name (and file path) to the goto file.

    Returns
    -------
    pd.DataFrame
        The route waypoints.
    """
    with open(filename, "r") as f:
        data = f.readlines()
    for i, line in enumerate(data):
        wpline = "b_arg: num_waypoints(nodim)"
        if line.lstrip().startswith("b_arg: num_waypoints(nodim)"):
            num_waypoints = int(line.lstrip().split(wpline)[1].split("#")[0])
        elif line.strip() == "<start:waypoints>":
            break
    waypoints = data[i + 1 : i + 1 + num_waypoints]
    waypoints = np.array([wp.split() for wp in waypoints])
    route = pd.DataFrame(
        {"longitude_text": waypoints[:, 0], "latitude_text": waypoints[:, 1]}
    )
    lon_spl = route.longitude_text.str.extract(r"([-]?)(\d+)(\d{2})\.(\d+)")
    route["longitude"] = np.array([1, -1])[(lon_spl[0] == "-").astype(int)] * (
        lon_spl[1].astype(float)
        + (lon_spl[2].astype(float) + lon_spl[3].astype(float) / 100) / 60
    )
    lat_spl = route.latitude_text.str.extract(r"([-]?)(\d+)(\d{2})\.(\d+)")
    route["latitude"] = np.array([1, -1])[(lat_spl[0] == "-").astype(int)] * (
        lat_spl[1].astype(float)
        + (lat_spl[2].astype(float) + lat_spl[3].astype(float) / 100) / 60
    )
    return route



from collections import namedtuple
import warnings
from matplotlib import pyplot as plt  # , patheffects as pe
import numpy as np
from scipy import interpolate
from sklearn import cluster
from . import meta, stats

Clustered = namedtuple("Clustered", ("x", "Y", "std", "std_unbiased", "count", "label"))


def add_credit(ax):
    """Add dreamcoat credit to axes.

    Parameters
    ----------
    ax : matplotlib.axes._subplots.AxesSubplot
        The axes to add the credit to.
    """
    ax.text(
        1.008,
        0.01,
        "dreamcoat {}".format(meta.version_colour),
        # "dreamcoat {} (v{})".format(meta.version_colour, meta.version_number),
        c="xkcd:{}".format(meta.version_colour),
        # fontweight="bold",
        # path_effects=[pe.withStroke(linewidth=0.8, alpha=0.7, foreground="#ceb301")],
        ha="left",
        va="bottom",
        rotation=-90,
        transform=ax.transAxes,
        fontsize=6.5,
    )


def get_clusters(x, Y, cluster_bandwidth):
    """_summary_

    Parameters
    ----------
    x : np.ndarray
        _description_
    Y : np.ndarray
        _description_
    cluster_bandwitdh : float
        The ``bandwith`` kwarg for ``sklearn.cluster.MeanShift``.

    Returns
    -------
    np.ndarray
        _description_
    np.ndarray
        _description_
    np.ndarray
        _description_
    """
    # Eliminate NaNs from input data
    l = ~np.isnan(x)
    x = x[l]
    Y = Y[l]
    # Do MeanShift clustering of the x-variable
    ms = cluster.MeanShift(bandwidth=cluster_bandwidth)
    ms.fit(np.vstack(x))
    # Get y-variable values at x-variable cluster centres
    cn = ms.labels_
    cx = ms.cluster_centers_.ravel()
    try:
        cY = np.full((cx.size, Y.shape[1]), np.nan)
        cY_std = np.full((cx.size, Y.shape[1]), np.nan)
        cY_std_unbiased = np.full((cx.size, Y.shape[1]), np.nan)
        cY_count = np.full((cx.size, Y.shape[1]), np.nan)
    except IndexError:
        cY = np.full(cx.size, np.nan)
        cY_std = np.full(cx.size, np.nan)
        cY_std_unbiased = np.full(cx.size, np.nan)
        cY_count = np.full(cx.size, 0)
    for i in range(cx.size):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="Degrees of freedom <= 0 for slice."
            )
            warnings.filterwarnings("ignore", message="Mean of empty slice")
            cY[i] = np.nanmean(Y[cn == i], axis=0)
            cY_std[i] = np.nanstd(Y[cn == i], axis=0)
        cY_std_unbiased[i] = stats.std_unbiased(Y[cn == i], axis=0)
        try:
            cY_count[i] = np.sum(np.vstack(cn == i) & ~np.isnan(Y), axis=0)
        except ValueError:
            cY_count[i] = np.sum((cn == i) & ~np.isnan(Y), axis=0)
    # Sort clusters
    ci = np.argsort(cx)
    cx = cx[ci]
    cY = cY[ci]
    cY_std = cY_std[ci]
    cY_std_unbiased = cY_std_unbiased[ci]
    cY_count = cY_count[ci]
    return Clustered(cx, cY, cY_std, cY_std_unbiased, cY_count, cn)


def get_cluster_profile(x, y, cluster_bandwidth=5, linspace_num=100):
    """Get clusters from input data and interpolate them."""
    cl = get_clusters(x, y, cluster_bandwidth=cluster_bandwidth)
    # Generate PCHIP interpolation of clustered data
    px = np.linspace(np.min(cl.x), np.max(cl.x), num=linspace_num)
    py = interpolate.pchip(cl.x, cl.Y)(px)
    return px, py, cl.x, cl.Y, cl.label


def cluster_profile(
    x,
    y,
    data=None,
    ax=None,
    cluster_bandwidth=5,
    linspace_num=100,
    invert_xy=False,
    plot_kwargs=None,
    scatter_kwargs=None,
):
    """Get clusters from input data, interpolate them and plot the results.

    Parameters
    ----------
    x : np.ndarray
        The independent variable, which is used for clustering.
    y : np.ndarray
        The dependent variable.
    data : dict-like, optional
        If provided, ``x`` and ``y`` are treated as being the keys (so the independent
        variable data is in ``data[x]`` and dependent in ``data[y]``).
    ax : optional
        Matplotlib axes to draw the figure on.  If not provided, new axes are created.
    cluster_bandwitdh : float, optional
        The ``bandwith`` kwarg for ``sklearn.cluster.MeanShift``, by default 5.
    linspace_num : int, optional
        The ``num`` kwarg for ``np.linspace`` interpolation, by default 100.
    invert_xy : bool, optional
        Whether to plot the independent variable on the x-axis (False) or y-axis (True),
        by default False.
    plot_kwargs : dict, optional
        Any kwargs to pass to the ``ax.plot`` artist.
    scatter_kwargs : dict, optional
        Any kwargs to pass to the ``ax.scatter`` artist.
    """
    # Extract data from data, if required
    if data is not None:
        x = data[x]
        y = data[y]
    p_xy = get_cluster_profile(
        x,
        y,
        cluster_bandwidth=cluster_bandwidth,
        linspace_num=linspace_num,
    )[:2]
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()
    if plot_kwargs is None:
        plot_kwargs = {}
    if scatter_kwargs is None:
        scatter_kwargs = {}
    xy = (x, y)
    if invert_xy:
        p_xy = p_xy[::-1]
        xy = xy[::-1]
    ax.plot(*p_xy, **plot_kwargs)
    ax.scatter(*xy, **scatter_kwargs)
    add_credit(ax)
    return fig, ax


def smooth_whittaker(y, factor=1, monotonic=True):
    """P-spline smoothing with the Whittaker graduation method.
    Modified from https://stats.stackexchange.com/a/467320

    Parameters
    ----------
    y : array-like
        The data to smooth, assumed to be evenly spaced.
    factor : float, optional
        Smoothing factor.  The default is 1.
    monotonic : bool, optional
        Whether the smoothed curve should increase monotonically. The default is True.

    Returns
    -------
    z : array-like
        The smoothed data, with the same shape as x and y.
    """
    L = ~np.isnan(y)
    # Prepare bases (Imat) and penalty
    dd = 3
    E = np.eye(L.sum())
    D3 = np.diff(E, n=dd, axis=0)
    D1 = np.diff(E, n=1, axis=0)
    kp = 1e12
    # Monotone smoothing
    ws = np.zeros(L.sum() - 1)
    for it in range(100):
        Ws = np.diag(ws * kp)
        zL = np.linalg.solve(E + factor * D3.T @ D3 + D1.T @ Ws @ D1, y[L])
        ws_new = D1 @ zL < 0.0
        dw = np.sum(ws != ws_new)
        ws = ws_new
        if dw == 0:
            break
    if monotonic:
        # Enforce true monotonicity
        zLd = np.diff(zL)
        for i in range(1, len(zL)):
            if zL[i] <= zL[i - 1]:
                zL[i] = zL[i - 1] + zL[i - 1] * 1e-12
    else:
        # Non-monotonic fit
        zL = np.linalg.solve(E + factor * D3.T @ D3, y[L])
    z = np.full(y.shape, np.nan)
    z[L] = zL
    return z



import warnings


def _error_check_point(point, correct_point=False):
    if len(point) != 2:
        raise TypeError(f"Point {str(point)} is incorrect length!")
    lon, lat = float(point[0]), float(point[1])
    if -90 <= lat <= 90 and -180 <= lon <= 180:  # Point makes sense
        return (lon, lat)
    elif -90 <= lon <= 90 and -180 <= lat <= 180:  # The point is (probably!) reversed
        _msg = (
            f"Point {str(point)} is probably reversed!\n\n"
            f"We believe that this is the case because\n"
            f"the provided longitude ({lon}) is in the range of [-90, 90]\n"
            f"and the provided latitude ({lat}) is in the range of [-180, 180].\n\n"
            f"The pont must be provided as a tuple of length 2, where the\n"
            f"first element is the longitude in the range of [-180, 180], and\n"
            f"the second element is the latitude in the range of [-90, 90].\n"
        )

        if correct_point:
            warnings.warn(_msg)
            warnings.warn(
                f"The parameter 'correct_point' has been set to True,\n"
                f"Therefore we are reversing the point and continuing..."
            )
            point_corrected = (lat, lon)
            return point_corrected
        else:
            raise ValueError(_msg)
    else:
        raise ValueError(
            f"\n"
            f"Point {str(point)} cannot be appropriately interpreted in this program.\n"
            f"Be advised, the pont must be provided as a tuple of length 2, where the\n"
            f"first element is the longitude in the range of [-180, 180], and\n"
            f"the second element is the latitude in the range of [-90, 90]).\n"
        )



"""
dreamcoat.underway
==================
Read and process underway data files (specifically, from RV Pelagia).

Functions
---------
read_sql
    Import the data from an sql underway file and parse it into a usable format.
"""

import pandas as pd
import numpy as np
from matplotlib import dates as mdates


def read_sql(filename):
    """Import the data from an sql underway file and parse it into a usable format.

    Parameters
    ----------
    filename : str
        The sql file name (and path).

    Returns
    -------
    pd.DataFrame
        The underway dataset.
    """
    # Import and parse underway SQL file
    with open(filename, "r") as f:
        underway_raw = f.readlines()
    headers = []
    is_header = False
    underway = []
    for line in underway_raw:
        if is_header:
            try:
                headers.append(line.split("`")[1])
            except IndexError:
                is_header = False
        if line.startswith("CREATE TABLE `aggregated` ("):
            is_header = True
        data_line = "INSERT INTO `aggregated` VALUES ("
        if line.startswith(data_line):
            line_data = (
                line.replace("'", "").split(data_line)[1].split(");\n")[0].split("),(")
            )
            line_data = [d.split(",") for d in line_data]
            underway = [*underway, *line_data]
    underway_array = np.array(underway)
    underway = pd.DataFrame({h: underway_array[:, i] for i, h in enumerate(headers)})
    # Convert all float columns to floats
    underway = underway.replace(["INVALID", ""], "-999")
    underway = underway.replace("°", "")
    for c in underway.columns:
        try:
            underway[c] = underway[c].astype(float)
            underway.loc[underway[c] == -999, c] = np.nan
        except ValueError:
            pass
    # Pythonise column names
    underway.columns = [
        c.lower().replace(" ", "_").replace("(", "_").replace(")", "")
        for c in underway.columns
    ]
    # Reformat where required
    underway["datetime"] = pd.to_datetime(underway.date)
    underway["datenum"] = mdates.date2num(underway.datetime)
    underway["latitude"] = underway.latitude.where(
        underway.latitude_direction == "N", -underway.latitude
    )
    underway["longitude"] = underway.longitude.where(
        underway.longitude_direction == "E", -underway.longitude
    )
    # Drop unnecessary columns
    underway = underway.drop(
        columns=[
            "date",
            "latitude_direction",
            "longitude_direction",
            "year",
            "month",
            "day",
        ]
    )
    return underway



"""
dreamcoat.maps
==============
Geographic mapping and route planning.

Functions
---------
build_vptree
    Build vantage-point tree for Natural Earth Data coastline.
get_distance
    Calculate distance between a pair of points.
get_distance_gcc
    Calculate distance between a pair of points in km using great_circle_calculator.
get_distance_geodesic
    Calculate geodesic distance between a single pair of points in km.
get_route_distance
    Calculate distance from point to point along a route.
linspace_gc
    Return evenly spaced points along a great circle.
linspace_gc_waypoints
    Apply `linspace_gc` across a series of waypoints, aiming for evenly spaced points.
linspace_gc_waypoints_cs
    Apply `linspace_gc` across a series of waypoints, with a constant number of points
    between each pair.
map_point_to_route
    Find the closest point on a route to a new point.
map_point_to_section
    Find the closest point on a single line section to a given point.
"""

import itertools
from collections import namedtuple
import numpy as np
from geographiclib.geodesic import Geodesic
from .. import gcc
from .vptree import VPTree
import cartopy.io.shapereader as shpreader
from shapely import geometry
from . import degrees_decimal_minutes


def linspace_gc(start, stop, num=50, endpoint=True):
    """Return evenly spaced points along a great circle.

    Parameters
    ----------
    start : (float, float)
        The (longitude, latitude) of the start point.
    stop : (float, float)
        The (longitude, latitude) of the end point.
    num : int, optional
        Number of points including start and stop, by default 50.
    endpoint : bool, optional
        Whether to include the endpoint, by default True.

    Returns
    -------
    (array_like, array_like)
        The (longitude, latitude) of the interpolated route.
    """
    wp_interp = np.full((num - (not endpoint), 2), np.nan)
    wp_interp[0] = start
    if endpoint:
        wp_interp[-1] = stop
    for i in range(1, num - 1):
        wp_interp[i] = gcc.intermediate_point(start, stop, fraction=i / (num - 1))
    return wp_interp.T


def linspace_gc_waypoints(waypoints, num_approx=100, endpoint=True):
    """Apply `linspace_gc()` across a series of waypoints, aiming for approximately
    `num_approx` roughly evenly spaced final points.

    Parameters
    ----------
    waypoints : (array_like, array_like)
        The (longitude, latitude) of the waypoints.
    num_approx : int, optional
        Approximate total number of points in the final interpolation, by default 100.
    endpoint : bool, optional
        Whether to include the final endpoint, by default True.

    Returns
    -------
    (array_like, array_like)
        The (longitude, latitude) of the interpolated waypoints.
    """
    # Calculate relative distance for each route step
    r_distance = get_route_distance(waypoints)
    r_fraction = np.diff(r_distance) / r_distance[-1]
    # Divide num_approx into the steps
    r_pts = (num_approx - 1) * r_fraction
    r_pts_round = np.round(r_pts).astype(int)
    r_zero = r_pts_round == 0
    r_pts_round[r_zero] += 1
    total_points = np.sum(r_pts_round) + 1
    # Do individual section interpolations and concatenate them
    wp_interp = []
    for i, npts in enumerate(r_pts_round):
        wp_interp.append(
            linspace_gc(
                waypoints[:, i],
                waypoints[:, i + 1],
                num=npts + 1,
                endpoint=i == len(r_pts_round) - 1 and endpoint,
            )
        )
    wp_interp = np.concatenate(wp_interp, axis=1)
    return wp_interp


def linspace_gc_waypoints_cs(waypoints, num_per_segment=50, endpoint=True):
    """Apply `linspace_gc()` across a series of waypoints with a constant number
    of interpolated points between each pair of waypoints.

    Parameters
    ----------
    waypoints : (array_like, array_like)
        The (longitude, latitude) of the waypoints.
    num_per_segment : int, optional
        Number of points between each pair of waypoints, by default 50.
    endpoint : bool, optional
        Whether to include the final endpoint, by default True.

    Returns
    -------
    (array_like, array_like)
        The (longitude, latitude) of the interpolated route.
    """
    wp_interp = []
    for i in range(1, route.shape[0]):
        wp_interp.append(
            linspace_gc(
                route[:, i - 1],
                route[:, i],
                num=num_per_segment,
                endpoint=i == route.shape[0] - 1 and endpoint,
            )
        )
    wp_interp = np.concatenate(wp_interp, axis=1)
    return wp_interp


def get_distance_geodesic(lon_lat_1, lon_lat_2):
    """Calculate geodesic distance between a single pair of points in km, based on
    https://stackoverflow.com/a/45480555

    Parameters
    ----------
    lon_lat_1 : array_like
        (longitude, latitude) of the first point in decimal degrees.
    lon_lat_2 : array_like
        (longitude, latitude) of the second point in decimal degrees.

    Returns
    -------
    float
        Distance between the two points in km.
    """
    return (
        Geodesic.WGS84.Inverse(lon_lat_1[1], lon_lat_1[0], lon_lat_2[1], lon_lat_2[0])[
            "s12"
        ]
        / 1e3
    )


def get_distance_gcc(lon_lat_1, lon_lat_2):
    """Calculate distance between a pair of points in km using great_circle_calculator.

    Parameters
    ----------
    lon_lat_1 : array_like
        (longitude, latitude) of the first point in decimal degrees.
    lon_lat_2 : array_like
        (longitude, latitude) of the second point in decimal degrees.

    Returns
    -------
    float
        Distance between the two points in km.
    """
    return gcc.distance_between_points(lon_lat_1, lon_lat_2, unit="kilometers")


def _get_distance_func(method):
    assert method in ["gcc", "geodesic"], "`method` must be either 'gcc' or 'geodesic'"
    if method == "gcc":
        return get_distance_gcc
    elif method == "geodesic":
        return get_distance_geodesic


def get_distance(lon_lat_1, lon_lat_2, method="gcc"):
    """Calculate distance between a pair of points in km using either
    great_circle_calculator (default) or geographiclib (more accurate but ~50 times
    slower).

    Parameters
    ----------
    lon_lat_1 : array_like
        (longitude, latitude) of the first point in decimal degrees.
    lon_lat_2 : array_like
        (longitude, latitude) of the second point in decimal degrees.
    method : str, optional
        Which method to use: 'gcc' (default) or 'geodesic'.

    Returns
    -------
    float
        Distance between the two points in km.
    """
    if isinstance(lon_lat_1, degrees_decimal_minutes.LatLon):
        lon_lat_1 = (lon_lat_1.longitude_dd, lon_lat_1.latitude_dd)
    if isinstance(lon_lat_2, degrees_decimal_minutes.LatLon):
        lon_lat_2 = (lon_lat_2.longitude_dd, lon_lat_2.latitude_dd)
    return _get_distance_func(method)(lon_lat_1, lon_lat_2)


def get_heading(lon_lat_1, lon_lat_2):
    """Calculate heading from lon_lat_1 to lon_lat_2.

    Parameters
    ----------
    lon_lat_1 : array_like
        (longitude, latitude) of the first point in decimal degrees.
    lon_lat_2 : array_like
        (longitude, latitude) of the second point in decimal degrees.

    Returns
    -------
    float
        Heading in ° (with N as 0, E as 90, ...).
    """
    if isinstance(lon_lat_1, degrees_decimal_minutes.LatLon):
        lon_lat_1 = (lon_lat_1.longitude_dd, lon_lat_1.latitude_dd)
    if isinstance(lon_lat_2, degrees_decimal_minutes.LatLon):
        lon_lat_2 = (lon_lat_2.longitude_dd, lon_lat_2.latitude_dd)
    heading = gcc.bearing_at_p1(lon_lat_1, lon_lat_2)
    if heading < 0:
        heading += 360
    return heading


def get_route_distance(waypoints, method="gcc"):
    """Calculate distance from point to point along a route.

    Parameters
    ----------
    waypoints : array_like
        The route (longitude, latitude) waypoints as a size (2, n) np.array.
    method : str, optional
        Which method to use: 'gcc' (default) or 'geodesic'.

    Returns
    -------
    array_like
        Distance from the start of the route in km.
    """
    route_distance = np.full_like(waypoints[0], 0)
    for i in range(1, len(route_distance)):
        route_distance[i] = (
            get_distance(waypoints[:, i - 1], waypoints[:, i], method=method)
            + route_distance[i - 1]
        )
    return route_distance


SectionPoint = namedtuple(
    "SectionPoint",
    ("distance_on_section", "distance_perpendicular", "lon_lat", "in_section"),
)
RoutePoint = namedtuple(
    "RoutePoint",
    ("distance_on_route", "lon_lat", "in_route", "waypoint_nearest"),
)


def map_point_to_section(lon_lat_1, lon_lat_2, lon_lat_pt, method="gcc"):
    """Find the closest point on a line section defined by `lon_lat_1` and `lon_lat_2`
    to a given point `lon_lat_pt` and calculate various associated properties.

    Parameters
    ----------
    lon_lat_1 : array_like
        (longitude, latitude) of the first point on the section in decimal degrees.
    lon_lat_2 : array_like
        (longitude, latitude) of the second point on the section in decimal degrees.
    lon_lat_pt : array_like
        (longitude, latitude) of the off-section point in decimal degrees.
    method : str, optional
        Which method to use for calculating distances: 'gcc' (default) or 'geodesic'.

    Returns
    -------
    namedtuple (SectionPoint) containing fields
        distance_on_section : float
            Distance between `lon_lat_1` and the mapping of `lon_lat_pt` onto the
            section in km.
        distance_perpendicular : float
            Distance between `lon_lat_pt` and its mapping onto the section in km.
        lon_lat : (float, float)
            (longitude, latitude) of the mapping of `lon_lat_pt` onto the section in
            decimal degrees.
        in_section : bool
            Whether the mapping of `lon_lat_pt` falls within the section.
    """
    distance_func = _get_distance_func(method)
    bearing_section = np.deg2rad(gcc.bearing_at_p1(lon_lat_1, lon_lat_2))
    bearing_point = np.deg2rad(gcc.bearing_at_p1(lon_lat_1, lon_lat_pt))
    angle_to_section = bearing_point - bearing_section
    distance_r0_pt = distance_func(lon_lat_1, lon_lat_pt)
    section_distance = distance_func(lon_lat_1, lon_lat_2)
    section_distance_pt = distance_r0_pt * np.sin(np.deg2rad(90) - angle_to_section)
    lon_lat_pt_section = gcc.point_given_start_and_bearing(
        lon_lat_1, np.rad2deg(bearing_section), section_distance_pt, unit="kilometers"
    )
    in_section = section_distance_pt >= 0 and section_distance_pt <= section_distance
    distance_perpendicular = distance_func(lon_lat_pt, lon_lat_pt_section)
    return SectionPoint(
        section_distance_pt,
        distance_perpendicular,
        lon_lat_pt_section,
        in_section,
    )


def map_point_to_route(route, lon_lat, extrapolate=False, verbose=False):
    """Find the closest point on a `route` to a new `lon_lat` point.

    Parameters
    ----------
    route : Route
        A Route object containing the route's waypoints.
    lon_lat : array_like
        (longitude, latitude) of the new off-route point.
    extrapolate : bool, optional
        Whether to extrapolate beyond the ends of the route, by default False.
    verbose : bool, optional
        Whether to report on progress, by default False.

    Results
    -------
    namedtuple (RoutePoint) containing fields
        distance_on_route : float
            Distance along the route up to the mapping of `lon_lat_pt` onto the route
            in km.
        lon_lat : (float, float)
            (longitude, latitude) of the mapping of `lon_lat` onto the route in decimal
            degrees.
        in_route : bool
            Whether the mapping of `lon_lat` falls within the route.
        waypoint_nearest : int
            The index of the nearest waypoint in `route.waypoints`.
    """

    def printv(*args, **kwargs):
        if verbose:
            print(*args, **kwargs)

    nearest = route.find_nearest_waypoint(lon_lat)
    section = None
    # Get the next section after the nearest waypoint, if possible
    section_next = None
    if nearest < route.size - 1:
        printv("Nearest waypoint is not the last waypoint.")
        section_next = map_point_to_section(
            route.waypoints[:, nearest],
            route.waypoints[:, nearest + 1],
            lon_lat,
            method=route.distance_method,
        )
    # Get the section directly before the nearest waypoint, if possible
    section_prev = None
    if nearest > 0:
        printv("Nearest waypoint is not the first waypoint.")
        section_prev = map_point_to_section(
            route.waypoints[:, nearest],
            route.waypoints[:, nearest - 1],
            lon_lat,
            method=route.distance_method,
        )
        # If using previous section, distance needs to be negative
        section_prev = SectionPoint(
            -section_prev.distance_on_section,
            section_prev.distance_perpendicular,
            section_prev.lon_lat,
            section_prev.in_section,
        )
    # Decide which of the two sections to use
    if section_next is not None and section_prev is not None:
        printv("Nearest waypoint is neither the first nor the last waypoint.")
        if section_next.in_section and section_prev.in_section:
            printv(
                " Point falls within both sections either side of the nearest waypoint."
            )
            if (
                section_next.distance_perpendicular
                <= section_prev.distance_perpendicular
            ):
                printv("  Point is closer to the section after the nearest waypoint.")
                section = section_next
            else:
                printv("  Point is closer to the section before the nearest waypoint.")
                section = section_prev
        elif section_next.in_section:
            printv(" Point falls only in the section after the nearest waypoint.")
            section = section_next
        elif section_prev.in_section:
            printv(" Point falls only in the section before the nearest waypoint.")
            section = section_prev
        else:
            printv(" Point does not fall in either section.")
            section = SectionPoint(0, 0, route.waypoints[:, nearest], True)
    elif section_next is not None:
        printv("Nearest waypoint is the first waypoint.")
        if section_next.in_section:
            printv(" Point falls within the first section.")
            section = section_next
        else:
            printv(" Point does not fall within the first section.")
            if extrapolate:
                printv("  Extrapolating before the first section.")
                section = section_next
            else:
                printv("  Returning the first waypoint.")
                section = SectionPoint(0, 0, route.waypoints[:, nearest], True)
    elif section_prev is not None:
        printv("Nearest waypoint is the last waypoint.")
        if section_prev.in_section:
            printv(" Point falls within the last section.")
            section = section_prev
        else:
            printv(" Point does not fall within the last section.")
            if extrapolate:
                printv("  Extrapolating after the last section.")
                section = section_prev
            else:
                printv("  Returning the last waypoint.")
                section = SectionPoint(0, 0, route.waypoints[:, nearest], True)
    # Convert section properties to route properties
    route_point = RoutePoint(
        route.distance[nearest] + section.distance_on_section,
        section.lon_lat,
        section.in_section,
        nearest,
    )
    return route_point


class Route:
    """A set of waypoints defining a route, which is assumed to travel in straight lines
    (i.e., along great circles) between the waypoints.

    Parameters
    ----------
    waypoints : np.ndarray
        A (2, n) NumPy array of the (longitude, latitude) values of the waypoints.
    distance_method : str, optional
        Which method is used to calculate distances, either
          "gcc" (default) - uses great_circle_calculator
          "geodesic" - uses geographiclib.geodesic.Geodesic.WGS84
        The first is an order of magnitude faster but a bit less accurate.

    Attributes
    ----------
    waypoints : np.ndarray
        The (2, n) NumPy array of the (longitude, latitude) values of the waypoints.
    size : int
        The number of waypoints (i.e., n).
    distance_method : str
        The method being used to calculate distances ("gcc" or "geodesic").
    distance : np.ndarray
        The distance of each waypoint along the route in km.
    vpt : vptree.VPTree
        A vantage-point tree for the route.
    wp_interp : np.ndarray
        A (2, n) NumPy array of the (longitude, latitude) values of a higher-resolution
        set of points interpolated between the waypoints.  Appears after running the
        interp method.
    wp_distance : np.ndarray
        The distance of each interpolated point in wp_interp along the route in km.
        Appears after running the interp method.

    Methods
    -------
    build_vpt
        Create a vantage-point tree for the route.
    find_nearest_waypoint
        Find the index of the waypoint closest to a given new point.
    get_distance
        Find the distance along the route corresponding to the closest point on the
        route to a new point.
    get_lon_lat
        Calculate the longitude and latitude of a given distance along the route.
    interp
        Interpolate the route to a higher resolution.
    map_point_to_route
        Find the closest point on the route to a new point.
    """

    def __init__(self, waypoints, distance_method="gcc"):
        assert (
            isinstance(waypoints, np.ndarray)
            and waypoints.shape[0] == 2
            and waypoints.ndim == 2
        ), "`waypoints` must be a numpy array of shape (2, n)."
        assert np.all(
            (waypoints[0] >= -180) & (waypoints[0] <= 180)
        ), "All longitudes must be from -180 to 180."
        assert np.all(
            (waypoints[1] >= -90) & (waypoints[1] <= 90)
        ), "All latitudes must be from -90 to 90."
        self.waypoints = waypoints
        self.size = waypoints.shape[1]
        self.distance_method = distance_method
        self.distance = get_route_distance(waypoints, method=distance_method)
        self.build_vpt()

    def build_vpt(self):
        """Create a vantage-point tree that can be used to find nearest neighbours."""
        self.vpt = VPTree(self.waypoints.T, _get_distance_func(self.distance_method))

    def find_nearest_waypoint(self, lon_lat):
        """Find the index of the waypoint closest to a new point.

        Parameters
        ----------
        lon_lat : array-like
            A two-element list or array containing the (longitude, latitude) of a new
            point, not necessarily on the route.

        Returns
        -------
        int
            The index of the waypoint closest to the new point.
        """
        nearest = self.vpt.get_nearest_neighbor(lon_lat)[1]
        return (
            (self.waypoints[0] == nearest[0]) & (self.waypoints[1] == nearest[1])
        ).argmax()

    def map_point_to_route(self, lon_lat, extrapolate=False, verbose=False):
        """Find the closest point on the route to a new point.

        Parameters
        ----------
        lon_lat : array-like
            (longitude, latitude) of the new off-route point.
        extrapolate : bool, optional
            Whether to extrapolate beyond the ends of the route, by default False.
        verbose : bool, optional
            Whether to report on progress, by default False.

        Results
        -------
        namedtuple (RoutePoint) containing fields
            distance_on_route : float
                Distance along the route up to the mapping of `lon_lat` onto the route
                in km.
            lon_lat : (float, float)
                (longitude, latitude) of the mapping of `lon_lat` onto the route in
                decimal degrees.
            in_route : bool
                Whether the mapping of `lon_lat` falls within the route.
            waypoint_nearest : int
                The index of the nearest waypoint.
        """
        return map_point_to_route(
            self, lon_lat, extrapolate=extrapolate, verbose=verbose
        )

    def interp(self, num_approx=100, endpoint=True):
        """Interpolate the route to a higher resolution, with the interpolated route
        stored in the attribute `wp_interp` and corresponding route distances in
        `wp_distance`.

        The interpolation runs such that the exact positions of the original waypoints
        are still included and the interpolated points are distributed as evenly as
        possible between the waypoints.

        Parameters
        ----------
        num_approx : int, optional
            Approximate total number of interpolated points, by default 100.
        endpoint : bool, optional
            Whether to include the very final waypoint in the interpolated set, by
            default True.
        """
        self.wp_interp = linspace_gc_waypoints(
            self.waypoints, num_approx=num_approx, endpoint=endpoint
        )
        self.wp_distance = get_route_distance(
            self.wp_interp, method=self.distance_method
        )

    def get_lon_lat(self, distance, extrapolate=False):
        """Get the longitude and latitude of a given distance along the route.

        Parameters
        ----------
        distance : float
            A distance along the route in km.
        extrapolate : bool, optional
            Whether to extrapolate beyond the ends of the route, if necessary; by
            default False.

        array-like
            (longitude, latitude) of the point along the route at the specified
            distance.
        """
        out_of_range = False
        if distance in self.distance:
            nearest = (self.distance == distance).argmax()
            lon_lat = self.waypoints[:, nearest]
        else:
            if distance < 0:
                after = 1
                before = 0
                out_of_range = True
            elif distance > self.distance.max():
                after = -1
                before = -2
                out_of_range = True
            else:
                after_point = distance < self.distance
                after = after_point.argmax()
                before = after - 1
            if not extrapolate and out_of_range:
                if distance < 0:
                    lon_lat = self.waypoints[:, 0]
                elif distance > self.distance.max():
                    lon_lat = self.waypoints[:, -1]
            else:
                section_fraction = (distance - self.distance[before]) / (
                    self.distance[after] - self.distance[before]
                )
                lon_lat = gcc.intermediate_point(
                    self.waypoints[:, before],
                    self.waypoints[:, after],
                    fraction=section_fraction,
                )
        return lon_lat

    def get_distance(self, lon_lat, extrapolate=False, verbose=False):
        """Get the route distance corresponding to a given longitude and latitude.

        Parameters
        ----------
        lon_lat : array-like
            (longitude, latitude) of the new off-route point.
        extrapolate : bool, optional
            Whether to extrapolate beyond the ends of the route, by default False.
        verbose : bool, optional
            Whether to report on progress, by default False.

        Returns
        -------
        float
            Distance along the route in km.
        """
        rp = self.map_point_to_route(lon_lat, extrapolate=extrapolate, verbose=verbose)
        return rp.distance_on_route


def geodesic_distance(lon_lat_1, lon_lat_2):
    """Calculate geodesic distance between a single pair of points in km, based on
    https://stackoverflow.com/a/45480555

    Parameters
    ----------
    lon_lat_1 : array_like
        (longitude, latitude) of the first point in decimal degrees.
    lon_lat_2 : array_like
        (longitude, latitude) of the second point in decimal degrees.

    Returns
    -------
    float
        Distance between the two points in km.
    """
    return (
        Geodesic.WGS84.Inverse(lon_lat_1[1], lon_lat_1[0], lon_lat_2[1], lon_lat_2[0])[
            "s12"
        ]
        / 1000
    )


def coastline_coords(lat_range, lon_range, resolution="10m"):
    """Get list of coastline coordinates from Natural Earth Data coastline sections
    that intersect with the lat_range and lon_range.

    Parameters
    ----------
    lat_range : array_like
        (min, max) latitude in decimal degrees.
    lon_range : array_like
        (min, max) longitude in decimal degrees.
    resolution : str, optional
        Natural Earth Data resolution, by default "10m" (can also be "50m" or "110m";
        default "10m" is the most detailed).

    Returns
    -------
    list
        Coordinates of the matching coastlines.
    """
    assert resolution in ["10m", "50m", "110m"]
    x = shpreader.natural_earth(
        resolution=resolution, category="physical", name="coastline"
    )
    cl = shpreader.Reader(x)
    coastlines = cl.records()
    my_bounds = geometry.Polygon(
        [
            (lon_range[0], lat_range[0]),
            (lon_range[0], lat_range[1]),
            (lon_range[1], lat_range[1]),
            (lon_range[1], lat_range[0]),
        ]
    )
    in_bounds = [cl for cl in coastlines if cl.geometry.intersects(my_bounds)]
    coords = list(itertools.chain(*[ib.geometry.coords for ib in in_bounds]))
    return coords


def build_vptree(lat_range, lon_range, **kwargs):
    """Build vantage-point tree from Natural Earth Data coastline sections that
    intersect with the lat_range and lon_range.

    Once constructed, use `vpt.get_nearest_neighbor((lon, lat))` to find the geodesic
    distance of nearest point on the coastline to (lon, lat) in km.

    Based on https://stackoverflow.com/a/45480555

    Parameters
    ----------
    lat_range : array_like
        (min, max) latitude in decimal degrees.
    lon_range : array_like
        (min, max) longitude in decimal degrees.

    Returns
    -------
    vpt : vptree.VPTree
        Vantage-point tree containing the matching coastline sections.
    """
    coords = coastline_coords(lat_range, lon_range, **kwargs)
    return VPTree(coords, geodesic_distance)


def extend_route(route_lon, route_lat, extra_fraction=0.05):
    """Add extra points at the start and end of a route.

    Parameters
    ----------
    route_lon : (n,) array-like
        The longitude points for the route in decimal degrees.
    route_lat : (n,) array-like
        The latitude points for the route in decimal degrees.
    extra_fraction : float, optional
        What fraction of the total route distance to extend by, by default 0.05.

    Returns
    -------
    route_lon_ext : (n + 2,) array-like
        The longitude points for the extended route in decimal degrees.
    route_lat_ext : (n + 2,) array-like
        The latitude points for the extended route in decimal degrees.
    route_distance_ext : (n + 2,) array-like
        The distances along the extended route in km from the first point of the
        original, unextended route.
    """
    route_distance = get_route_distance(np.array([route_lon, route_lat]))
    total_distance = route_distance[-1]
    bearing_start = gcc.bearing_at_p1(
        (route_lon[1], route_lat[1]), (route_lon[0], route_lat[0])
    )
    bearing_end = gcc.bearing_at_p1(
        (route_lon[-2], route_lat[-2]), (route_lon[-1], route_lat[-1])
    )
    extra_start = gcc.point_given_start_and_bearing(
        (route_lon[0], route_lat[0]),
        bearing_start,
        extra_fraction * total_distance * 1000,
    )
    extra_end = gcc.point_given_start_and_bearing(
        (route_lon[-1], route_lat[-1]),
        bearing_end,
        extra_fraction * total_distance * 1000,
    )
    route_lon_ext = np.array([extra_start[0], *route_lon, extra_end[0]])
    route_lat_ext = np.array([extra_start[1], *route_lat, extra_end[1]])
    route_distance_ext = get_route_distance(np.array([route_lon_ext, route_lat_ext]))
    route_distance_ext -= route_distance_ext[1]
    return route_lon_ext, route_lat_ext, route_distance_ext



import dreamcoat as dc
import numpy as np


def test_std_bias_correction():
    # Test values taken from table in section Bias correction
    # https://en.wikipedia.org/wiki/Unbiased_estimation_of_standard_deviation
    # on 4th May 2024
    sample_size = np.array([2, 3, 4, 5, 6, 7, 8, 9, 10, 100])
    c4 = dc.stats.std_bias_correction(sample_size)
    c4_wikipedia = np.array(
        [
            0.7978845608,  # 2
            0.8862269255,  # 3
            0.9213177319,  # 4
            0.9399856030,  # 5
            0.9515328619,  # 6
            0.9593687891,  # 7
            0.9650304561,  # 8
            0.9693106998,  # 9
            0.9726592741,  # 10
            0.9974779761,  # 100
            # Can't use sample_sizes below here because factorial(k - 1) returns inf
            # 0.9997497811,  # 1000
            # 0.9999749978,  # 10000
        ]
    )
    diff = np.round(c4, decimals=10) - c4_wikipedia
    # We get the final (10th) decimal place slightly wrong for sample_sizes 7 and 9
    assert np.all(np.abs(diff) < 1e-9)


# test_std_bias_correction()



import warnings
import numpy as np
from scipy.special import factorial


def std_bias_correction(sample_size):
    """Find bias correction for standard deviations computed from small numbers of
    normally distributed measurements.

    Parameters
    ----------
    sample_size : int
        Number of measurements from which the standard deviation was computed.

    Returns
    -------
    c4 : float
        The correction factor c4, based on the equations from
        https://en.wikipedia.org/wiki/Unbiased_estimation_of_standard_deviation
        The standard deviation obtained from np.std(a, ddof=1) should be divided by
        this c4 value to find the unbiased estimate.
    """
    k = np.floor(sample_size / 2)
    with np.errstate(divide="ignore", invalid="ignore"):
        c4 = np.where(
            sample_size % 2 == 0,
            # For even sample_size:
            np.sqrt(2 / (np.pi * (2 * k - 1)))
            * (2 ** (2 * k - 2) * factorial(k - 1) ** 2)
            / factorial(2 * k - 2),
            # For odd sample_size:
            np.sqrt(np.pi / k)
            * factorial(2 * k - 1)
            / (2 ** (2 * k - 1) * factorial(k - 1) ** 2),
        )
    return c4


def std_unbiased(a, axis=None, **kwargs):
    """Compute standard deviation unbiased for the number of samples, assuming a normal
    distribution.  The correction factor c4 is used, based on the equations from
    https://en.wikipedia.org/wiki/Unbiased_estimation_of_standard_deviation

    Parameters
    ----------
    a : array-like
        The values to find the standard deviation of.
    axis : int, optional
        Which axis to find the standard deviation along, by default None.  For example,
        with an array of shape (1000, 4) and axis=1, this represents 1000 different
        groups of 4 measurements.

    Returns
    -------
    float
        The unbiased standard deviation, corrected for the number of samples.
    """
    if axis is None:
        size_a = np.size(a)
    else:
        size_a = np.shape(a)[axis]
    c4 = std_bias_correction(size_a)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Degrees of freedom <= 0 for slice.")
        std_biased = np.nanstd(a, axis=axis, ddof=1, **kwargs)
        # ^ ddof must be 1 in order for the c4 correction to be valid
    return std_biased / c4



import networkx as nx
import numpy as np
from cartopy import crs as ccrs
from cartopy import feature as cfeature
from matplotlib import pyplot as plt
from scipy import interpolate

from . import maps, plot


class CruiseGraph(nx.Graph):
    """A graph representation of a cruise for computing neutral surfaces.

    Parameters
    ----------
    ctdz : pd.DataFrame
        CTD sensor data for a cruise binned at regular depth or pressure intervals and
        including only downcast data.  Must contain at least the following columns:
            station : int or str
                An identifier that is unique for each CTD station.
            latitude, longitude : float
                Latitude and longitude in decimal degrees (N and E positive).
            pressure : float
                Hydrostatic pressure in dbar.
            theta : float
                Potential temperature in °C.
            salinity : float
                Practical salinity.
        The ctdz dataframe is stored in CruiseGraph.ctdz.
    edges : array-like, optional
        List of adjacency edges between pairs of stations, in the format required by
        nx.Graph.add_edges_from(), where edges here = ebunch_to_add there.

    Methods
    -------
    get_surfaces
        Compute all omega surfaces relative to a given reference station.
    get_surfaces_all
        Compute all omega surfaces with every station used as the reference separately.
    plot_surface_map
        Draw the adjacency graph and pressure on a given surface.
    """

    def __init__(self, ctdz, edges=None):
        super().__init__()
        self.ctdz = ctdz.copy()
        self._get_veronis_here()
        self._get_stations()
        self.surfaces_raw = {}
        self.surfaces = {}
        if edges is not None:
            self.add_edges_from(edges)

    def _get_veronis_here(self):
        # Calculate Veronis densities with each station referenced to itself
        # from neutralocean.label import veronis  # for neutralocean v2.2.0
        from neutralocean.label import veronis_density  # for neutralocean v2.1.3

        self.ctdz["veronis_here_raw"] = np.nan
        self.ctdz["veronis_here"] = np.nan
        for s, station in self.ctdz.groupby("station"):
            for i, row in station.iterrows():
                # For neutralocean v2.1.3:
                self.ctdz.loc[i, "veronis_here_raw"] = veronis_density(
                    station.salinity.values,
                    station.theta.values,
                    station.pressure.values,
                    row.pressure,
                    eos="jmdfwg06",
                )
                # # For neutralocean v2.2.0:
                # self.ctdz.loc[i, "veronis_here_raw"] = veronis(
                #     row.pressure,
                #     station.salinity.values,
                #     station.theta.values,
                #     station.pressure.values,
                #     eos="jmdfwg06",
                # )
        self.ctdz["veronis_here_raw"] -= 1000
        for s, station in self.ctdz.groupby("station"):
            # TODO add remove outliers step before smoothing!  (search "cutoff" below)
            S = self.ctdz.station == s
            self.ctdz.loc[S, "veronis_here"] = plot.smooth_whittaker(
                station.veronis_here_raw.values
            )

    def _get_stations(self):
        self.stations = (
            self.ctdz[["station", "longitude", "latitude"]].groupby("station").mean()
        )
        self.stations["npts"] = (
            self.ctdz[["station", "longitude"]].groupby("station").count()
        )
        self.add_nodes_from(self.stations.index)

    def _get_edge_distances(self):
        distances = {}
        for _a, _b in self.edges:
            row_a = self.stations.loc[_a]
            row_b = self.stations.loc[_b]
            distances[(_a, _b)] = maps.get_distance_geodesic(
                (row_a.longitude, row_a.latitude),
                (row_b.longitude, row_b.latitude),
            )
        nx.set_edge_attributes(self, distances, "distance")

    def add_edges_from(self, ebunch_to_add, **attr):
        assert np.all(np.isin(ebunch_to_add, self.stations.index)), (
            "All nodes must be in the existing list of stations!"
        )
        super().add_edges_from(ebunch_to_add, **attr)
        self._get_edge_distances()
        self._get_grid()
        self._get_stp()

    def add_edge(self, u_of_edge, v_of_edge, **attr):
        assert u_of_edge in self.stations.index and v_of_edge in self.stations.index, (
            "Both u_of_edge and v_of_edge must be in the existing list of stations!"
        )
        super().add_edge(u_of_edge, v_of_edge, **attr)
        self._get_edge_distances()
        self._get_grid()
        self._get_stp()

    def _get_grid(self):
        a = []
        b = []
        distances = []
        for edge in self.edges:
            _a, _b = edge
            a.append((self.stations.index == _a).nonzero()[0][0])
            b.append((self.stations.index == _b).nonzero()[0][0])
            distances.append(self.edges[edge]["distance"])
        self.grid = {
            "edges": np.array([a, b]),
            "dist": np.array(distances),
            "distperp": np.ones_like(distances),
        }

    def _get_stp(self):
        stp_vars = ["salinity", "theta", "pressure"]
        stp = {}
        for v in stp_vars:
            stp[v] = np.full((self.stations.shape[0], self.stations.npts.max()), np.nan)
            for i, s in enumerate(self.stations.index):
                S = self.ctdz.station == s
                stp[v][i, : S.sum()] = self.ctdz[v][S].values
        self.stp = [stp[v] for v in stp_vars]

    def get_surfaces(self, station_ref, cutoff=5):
        """Compute all omega surfaces relative to a given reference station.

        After computing, surfaces are processed to remove outliers and smoothed so that
        each surface increases monotonically.  Veronis density labels are assigned from
        the reference station.

        Raw computed surfaces are stored in CruiseGraph.surfaces_raw[station_ref], and
        processed surfacs are in CruiseGraph.surfaces[station_ref].

        The processed surfaces and their labels are also added to CruiseGraph.ctdz in
        columns called "p_at_<station_ref>" and "veronis_<station_ref>".

        Parameters
        ----------
        station_ref : int or str
            The identifier for the station to use as the reference.
        cutoff : float
            The cutoff for outlier removal in dbar, by default 5.  Set to a very high
            number to not remove outliers.
        """
        from neutralocean.surface import omega_surf

        i_ref = self._station_to_index(station_ref)
        surfaces = np.full((self.stations.shape[0], self.stations.npts.max()), np.nan)
        for i, p_init in enumerate(
            self.ctdz[self.ctdz.station == station_ref].pressure
        ):
            # For neutralocean v2.1.3:
            surfaces[:, i] = omega_surf(
                *self.stp,
                self.grid,
                pin_cast=i_ref,
                pin_p=p_init,
                eos="jmdfwg06",
                output=False,
            )[2]
            # # For neutralocean v2.2.0:
            # surfaces[:, i] = omega_surf(
            #     *self.stp,
            #     self.grid,
            #     pin_cast=i_ref,
            #     p_init=p_init,
            #     eos="jmdfwg06",
            #     output=False
            # )[2]
        self.surfaces_raw[station_ref] = surfaces
        self._smooth_and_interpolate(station_ref, cutoff=cutoff)
        self._get_ref_surfaces(station_ref)
        self._label_surfaces(station_ref)

    def get_surfaces_all(self, cutoff=5, verbose=True):
        """Compute all omega surfaces using CruiseGraph.get_surfaces() with every
        station used as the reference separately.

        Parameters
        ----------
        cutoff : float
            The cutoff for outlier removal in dbar, by default 5.  Set to a very high
            number to not remove outliers.
        verbose : bool
            Whether to print progress, by default True.
        """
        for i, station_ref in enumerate(self.stations.index):
            if verbose:
                print(
                    "Getting surfaces wrt. station {} ({} of {})...".format(
                        station_ref, i + 1, self.stations.index.size
                    )
                )
            self.get_surfaces(station_ref, cutoff=cutoff)
        if verbose:
            print("Got all surfaces!")

    def _remove_outliers(self, station_ref, cutoff=5):
        surfaces_raw_clean = self.surfaces_raw[station_ref].copy()
        for i, station in enumerate(surfaces_raw_clean):
            for j in range(1, len(station) - 1):
                if (
                    (np.abs(station[j] - station[j - 1]) >= cutoff)
                    & (np.abs(station[j] - station[j + 1]) >= cutoff)
                    & (np.abs(station[j - 1] - station[j + 1]) <= cutoff)
                ):
                    surfaces_raw_clean[i, j] = (station[j - 1] + station[j + 1]) / 2
        return surfaces_raw_clean

    def _smooth_and_interpolate(self, station_ref, cutoff=5):
        surfaces_raw_clean = self._remove_outliers(station_ref, cutoff=cutoff)
        self.surfaces[station_ref] = np.full(surfaces_raw_clean.shape, np.nan)
        pressure_ref = self.get_pressure_ref(station_ref)
        for i in range(len(self.stations.index)):
            level = plot.smooth_whittaker(surfaces_raw_clean[i])
            # Remove NaNs for interpolation and interpolate
            L = ~np.isnan(level)
            interp = interpolate.PchipInterpolator(
                pressure_ref[L], level[L], extrapolate=False
            )
            self.surfaces[station_ref][i] = interp(pressure_ref)

    def _get_ref_surfaces(self, station_ref):
        pressure_ref = self.get_pressure_ref(station_ref)
        self.ctdz["p_at_{}".format(station_ref)] = np.nan
        for i, s in enumerate(self.stations.index):
            L = ~np.isnan(self.surfaces[station_ref][i])
            interp = interpolate.PchipInterpolator(
                self.surfaces[station_ref][i][L], pressure_ref[L], extrapolate=False
            )
            S = self.ctdz.station == s
            self.ctdz.loc[S, "p_at_{}".format(station_ref)] = interp(
                self.ctdz.loc[S, "pressure"]
            )

    def _label_surfaces(self, station_ref):
        ref_veronis = self.ctdz[self.ctdz.station == station_ref][
            ["pressure", "veronis_here"]
        ]
        interp_veronis = interpolate.PchipInterpolator(
            ref_veronis.pressure, ref_veronis.veronis_here
        )
        self.ctdz["veronis_{}".format(station_ref)] = interp_veronis(
            self.ctdz["p_at_{}".format(station_ref)]
        )

    def _station_to_index(self, station):
        return (self.stations.index == station).nonzero()[0][0]

    def _index_to_station(self, index):
        return self.stations.index[index]

    def get_pressure_ref(self, station_ref):
        i_ref = self._station_to_index(station_ref)
        pressure_ref = self.stp[2][i_ref]
        return pressure_ref

    def _get_pos(self, crs):
        self.crs = crs
        self.pos = {
            i: self.crs.transform_points(
                ccrs.PlateCarree(), row.longitude, row.latitude
            )[0][:2]
            for i, row in self.stations.iterrows()
        }

    def plot_surface_map(
        self,
        station_ref,
        p_surface,
        crs=None,
        cutoff=5,
        dpi=300,
        extent=None,
        pad_latitude=0.2,
        pad_longitude=0.2,
    ):
        """Draw the adjacency graph and pressure on a given surface.

        Parameters
        ----------
        station_ref : int or str
            The identifier for the station to use as the reference.  The method
            CruiseGraph.get_surfaces(station_ref) is run if it hasn't already been.
        p_surface : float
            What pressure at the reference station to plot the surface for.  Does not
            need to match exactly, the closest one will be used.
        crs : Cartopy crs
            A coordinate reference system from cartopy.crs.  If not provided, then
            cartopy.crs.PlateCarree() is used.
        cutoff : float, optional
            The cutoff for outlier removal in dbar, by default 5.  Set to a very high
            number to not remove outliers.  Only used if the method
            CruiseGraph.get_surfaces(station_ref) needs to be run.
        dpi : int, optional
            Figure resolution, by default 300.
        extent : array-like, optional
            Map extent, as [lon_min, lon_max, lat_min, lat_max].  If not provided,
            the extent of the stations plus padding is used (see pad_latitude and
            pad_longitude).
        pad_latitude : float, optional
            Fraction by which to pad the latitude axis if extent not provided, by
            default 0.2.
        pad_longitude : float, optional
            Fraction by which to pad the longitude axis if extent not provided, by
            default 0.2.
        """
        if station_ref not in self.surfaces:
            self.get_surfaces(station_ref, cutoff=cutoff)
        if extent is None:
            lat_min = self.stations.latitude.min()
            lat_max = self.stations.latitude.max()
            lon_min = self.stations.longitude.min()
            lon_max = self.stations.longitude.max()
            lat_diff = lat_max - lat_min
            lon_diff = lon_max - lon_min
            extent = [
                lon_min - lon_diff * pad_longitude,
                lon_max + lon_diff * pad_longitude,
                lat_min - lat_diff * pad_latitude,
                lat_max + lat_diff * pad_latitude,
            ]
        pressure_ref = self.get_pressure_ref(station_ref)
        i = np.argmin(np.abs(pressure_ref[~np.isnan(pressure_ref)] - p_surface))
        if i >= self.stations.loc[station_ref].npts:
            i = self.stations.loc[station_ref].npts - 1
        i = int(i)
        if crs is None:
            crs = ccrs.PlateCarree()
        self._get_pos(crs)
        fig, ax = plt.subplots(dpi=dpi, subplot_kw={"projection": crs})
        nx.draw_networkx(
            self,
            pos=self.pos,
            ax=ax,
            # Node marker properties
            node_size=0,
            linewidths=0,
            # Edge properties
            width=0.5,
            # Label properties
            font_size=8,
            with_labels=False,
        )
        sc = ax.scatter(
            "longitude",
            "latitude",
            data=self.stations,
            transform=ccrs.PlateCarree(),
            s=20,
            c=self.surfaces[station_ref][:, i],
            zorder=10,
            cmap="viridis_r",
            vmin=self.ctdz.loc[self.ctdz.station == station_ref].pressure.min(),
            vmax=self.ctdz.loc[self.ctdz.station == station_ref].pressure.max(),
        )
        plt.colorbar(sc, label="Neutral surface pressure / dbar")
        ax.scatter(
            "longitude",
            "latitude",
            data=self.stations.loc[station_ref],
            s=60,
            c="none",
            edgecolor="xkcd:strawberry",
            zorder=11,
            transform=ccrs.PlateCarree(),
        )
        ax.set_extent(extent, crs=ccrs.Geodetic())
        ax.set_title(
            "Ref. pressure = {:.1f} dbar".format(
                self.ctdz.loc[self.ctdz.station == station_ref].pressure.iloc[i]
            )
        )
        ax.add_feature(
            cfeature.NaturalEarthFeature(
                "physical", "land", "10m", edgecolor="none", facecolor="xkcd:dark"
            )
        )
        plot.add_credit(ax)
        fig.tight_layout()
        return fig, ax



from ._constants import *
from .__error_checking import _error_check_point


def _degrees_to_radians(degrees):
    """Converts the degrees into radians
    :param degrees: decimal degrees
    :return: radians
    """
    return pi * degrees / 180


def _radians_to_degrees(radians):
    """Converts the radians into degrees
    :param radians: decimal degrees
    :return: radians
    """
    return 180 * radians / pi


def _point_to_radians(point):
    point = _error_check_point(point)
    return (_degrees_to_radians(point[0]), _degrees_to_radians(point[1]))


def _point_to_degrees(point):
    return (_radians_to_degrees(point[0]), _radians_to_degrees(point[1]))



"""
Dreamcoat
=========

> The joyful oceanographic seagoing expedition planning helper

Dreamcoat is a set of tools to help work with oceanographic data of the sort that might
be needed or generated during a research expedition.

Dreamcoat is being developed by Dr Matthew P. Humphreys (https://hseao3.group) at the
NIOZ Royal Netherlands Institute for Sea Research, Texel.  You're very welcome to give
it a go and suggestions and contributions are welcome, but please be aware that it is
primarily designed as a personal tool so there probably won't be much help if you get
stuck; nothing is guaranteed not to break between versions, and the documentation
(https://dreamcoat.hseao3.group) may be incomplete or out of date.

The docstring examples assume that `dreamcoat` has been imported as ``dc``:

  >>> import dreamcoat as dc

Modules
-------
convert
    Convert between different units and formats.
ctd
    Dealing with CTD data files.
gcc
    The great_circle_calculator package.
glider
    Working with Slocum gliders.
maps
    Working with data in (longitude, latitude) space.
meta
    Metadata for dreamcoat.
neutral
    Working with neutralocean surfaces.
plot
    Shortcuts for data visualisations.
underway
    Working with underway data files.

Classes
-------
CruiseGraph
    A graph representation of a cruise for computing neutral surfaces.
LatLon
    A tool to convert between decimal degrees and degrees decimal minutes.
Route
    A set of waypoints defining a route.

Functions
---------
hello
    Report the version number.
"""

from .great_circle_calculator import great_circle_calculator as gcc
from . import convert, ctd, glider, maps, meta, neutral, plot, stats, underway
from .maps import Route
from .maps.degrees_decimal_minutes import LatLon
from .meta import __version__, hello
from .neutral import CruiseGraph



from .__conversion import (
    _point_to_radians,
    _point_to_degrees,
    _radians_to_degrees,
    _degrees_to_radians,
)
from .__error_checking import _error_check_point
from ._constants import *


def distance_between_points(p1, p2, unit="meters", haversine=True):
    """This function computes the distance between two points in the unit given in the unit parameter.  It will
    calculate the distance using the haversine unless the user specifies haversine to be False.  Then law of cosines
    will be used
    :param p1: tuple point of (lon, lat)
    :param p2: tuple point of (lon, lat)
    :param unit: unit of measurement. List can be found in constants.eligible_units
    :param haversine: True (default) uses haversine distance, False uses law of cosines
    :return: Distance between p1 and p2 in the units specified.
    """
    lon1, lat1 = _point_to_radians(_error_check_point(p1))
    lon2, lat2 = _point_to_radians(_error_check_point(p2))
    r_earth = getattr(radius_earth, unit, "meters")
    if haversine:
        # Haversine
        d_lat, d_lon = lat2 - lat1, lon2 - lon1
        a = sin(d_lat / 2) * sin(d_lat / 2) + cos(lat1) * cos(lat2) * sin(
            d_lon / 2
        ) * sin(d_lon / 2)
        a = min(1, a)
        c = 2 * atan2(sqrt(a), sqrt((1 - a)))
        dist = r_earth * c
        return dist
    # Spherical Law Of Cosines
    dist = (
        acos(sin(lat1) * sin(lat2) + cos(lat1) * cos(lat2) * cos(lon2 - lon1)) * r_earth
    )
    return dist


def bearing_at_p1(p1, p2):
    """This function computes the bearing (i.e. course) at p1 given a destination of p2.  Use in conjunction with
    midpoint(*) and intermediate_point(*) to find the course along the route.  Use bearing_at_p2(*) to find the bearing
    at the endpoint
    :param p1: tuple point of (lon, lat)
    :param p2: tuple point of (lon, lat)
    :return: Course, in degrees
    """
    lon1, lat1 = _point_to_radians(_error_check_point(p1))
    lon2, lat2 = _point_to_radians(_error_check_point(p2))
    x = cos(lat1) * sin(lat2) - sin(lat1) * cos(lat2) * cos(lon2 - lon1)
    y = sin(lon2 - lon1) * cos(lat2)
    course = atan2(y, x)
    return _radians_to_degrees(course)


def bearing_at_p2(p1, p2):
    """This function computes the bearing (i.e. course) at p2 given a starting point of p1.  Use in conjunction with
    midpoint(*) and intermediate_point(*) to find the course along the route.  Use bearing_at_p1(*) to find the bearing
    at the endpoint
    :param p1: tuple point of (lon, lat)
    :param p2: tuple point of (lon, lat)
    :return: Course, in degrees
    """
    return (bearing_at_p1(p2, p1) + 180) % 360


def midpoint(p1, p2):
    """This is the half-way point along a great circle path between the two points.
    :param p1: tuple point of (lon, lat)
    :param p2: tuple point of (lon, lat)
    :return: point (lon, lat)
    """
    lon1, lat1 = _point_to_radians(_error_check_point(p1))
    lon2, lat2 = _point_to_radians(_error_check_point(p2))
    b_x = cos(lat2) * cos(lon2 - lon1)
    b_y = cos(lat2) * sin(lon2 - lon1)
    lat3 = atan2(
        sin(lat1) + sin(lat2), sqrt((cos(lat1) + b_x) * (cos(lat1) + b_x) + b_y * b_y)
    )
    lon3 = lon1 + atan2(b_y, cos(lat1) + b_x)
    lat3 = _radians_to_degrees(lat3)
    lon3 = (_radians_to_degrees(lon3) + 540) % 360 - 180
    p3 = (lon3, lat3)
    return p3


def intermediate_point(p1, p2, fraction=0.5):
    """This function calculates the intermediate point along the course laid out by p1 to p2.  fraction is the fraction
    of the distance between p1 and p2, where 0 is p1, 0.5 is equivalent to midpoint(*), and 1 is p2.
    :param p1: tuple point of (lon, lat)
    :param p2: tuple point of (lon, lat)
    :param fraction: the fraction of the distance along the path.
    :return: point (lon, lat)
    """
    lon1, lat1 = _point_to_radians(_error_check_point(p1))
    lon2, lat2 = _point_to_radians(_error_check_point(p2))
    delta = distance_between_points(p1, p2) / radius_earth.meters
    a = sin((1 - fraction) * delta) / sin(delta)
    b = sin(fraction * delta) / sin(delta)
    x = a * cos(lat1) * cos(lon1) + b * cos(lat2) * cos(lon2)
    y = a * cos(lat1) * sin(lon1) + b * cos(lat2) * sin(lon2)
    z = a * sin(lat1) + b * sin(lat2)
    lat3 = atan2(z, sqrt(x * x + y * y))
    lon3 = atan2(y, x)
    return _point_to_degrees((lon3, lat3))


def point_given_start_and_bearing(p1, course, distance, unit="meters"):
    """Given a start point, initial bearing, and distance, this will calculate the destina­tion point and final
    bearing travelling along a (shortest distance) great circle arc.
    :param p1: tuple point of (lon, lat)
    :param course: Course, in degrees
    :param distance: a length in unit
    :param unit: unit of measurement. List can be found in constants.eligible_units
    :return: point (lon, lat)
    """
    lon1, lat1 = _point_to_radians(_error_check_point(p1))
    brng = _degrees_to_radians(course)
    r_earth = getattr(radius_earth, unit, "meters")
    delta = distance / r_earth
    lat2 = asin(sin(lat1) * cos(delta) + cos(lat1) * sin(delta) * cos(brng))
    lon2 = lon1 + atan2(
        sin(brng) * sin(delta) * cos(lat1), cos(delta) - sin(lat1) * sin(lat2)
    )
    lon2 = (_radians_to_degrees(lon2) + 540) % 360 - 180
    p2 = (lon2, _radians_to_degrees(lat2))
    return p2



"""
dreamcoat.convert
=================
Conversions between different units and formats.

Functions
---------
extent_to_nsew
    Get "N", "S", "E" and "W" from map extent values depending on their sign.
knots_to_kph
    Knots into kilometers per hour.
kph_to_knots
    Kilometers per hour into knots.
nm_to_km
    Nautical miles into kilometers.
km_to_nm
    Kilometers into nautical miles.
cartesian_to_polar
    Cartesian into polar coordinates.
polar_to_cartesian
    Polar into Cartesian coordinates.
"""

import numpy as np


def extent_to_nsew(map_extent):
    """Get "N", "S", "E" and "W" from map extent values depending on their sign.

    Parameters
    ----------
    map_extent : list of float
        [west, east, north, south] map extents in decimal degrees.

    Returns
    -------
    list of str
        ["E"/"W", "E"/"W", "N"/"S", "N"/"S"] as appropriate for the given extents.
    """
    nsew = [
        " EW"[int(np.sign(map_extent[0]))],
        " EW"[int(np.sign(map_extent[1]))],
        " NS"[int(np.sign(map_extent[2]))],
        " NS"[int(np.sign(map_extent[3]))],
    ]
    return nsew


def knots_to_kph(knots):
    """Convert speed in knots to kilometers per hour.

    Parameters
    ----------
    knots : float
        Speed in knots.

    Returns
    -------
    float
        Speed in km/h.
    """
    return knots * 1.852


def kph_to_knots(kph):
    """Convert speed in kilometers per hour to knots.

    Parameters
    ----------
    kph : float
        Speed in km/h.

    Returns
    -------
    float
        Speed in knots.
    """
    return kph / 1.852


def nm_to_km(nm):
    """Convert distance in nautical miles to kilometers.

    Parameters
    ----------
    nm : float
        Distance in nautical miles.

    Returns
    -------
    float
        Distance in km.
    """
    return nm * 1.852


def km_to_nm(km):
    """Convert distance in kilometers to nautical miles.

    Parameters
    ----------
    km : float
        Distance in kilometers.

    Returns
    -------
    float
        Distance in nautical miles.
    """
    return km / 1.852


def cartesian_to_polar(x, y):
    """Convert from Cartesian to polar co-ordinates, where theta is zero in the N
    position and increases clockwise.

    Parameters
    ----------
    x : float
        x-axis value in Cartesian co-ordinates.
    y : float
        y-axis value in Cartesian co-ordinates.

    Returns
    -------
    theta
        Angle in polar co-ordinates.
    rho
        Radius in polar co-ordinates.
    """
    rho = np.sqrt(x**2 + y**2)
    theta = np.arctan2(y, x)
    return theta, rho


def polar_to_cartesian(theta, rho):
    """Convert from polar to Cartesian co-ordinates, where theta is zero in the N
    position and increases clockwise.

    Parameters
    ----------
    theta : float
        Angle in polar co-ordinates.
    rho : float
        Radius in polar co-ordinates.

    Returns
    -------
    x : float
        x-axis value in Cartesian co-ordinates.
    y : float
        y-axis value in Cartesian co-ordinates.
    """
    x = rho * np.cos(theta)
    y = rho * np.sin(theta)
    return x, y


def spherical_to_cartesian(azimuth, elevation, r):
    """Convert from spherical to Cartesian co-ordinates, following MATLAB's sph2cart.

    Parameters
    ----------
    azimuth : float
        The azimuth in degrees.
    elevation : float
        The elevation in degrees.
    r : float
        The radius.

    Returns
    -------
    x, y, z : floats
        Cartesian coordinates in the x, y and z dimensions.
    """
    x = r * np.cos(elevation) * np.cos(azimuth)
    y = r * np.cos(elevation) * np.sin(azimuth)
    z = r * np.sin(elevation)
    return x, y, z


def cartesian_to_spherical(x, y, z):
    """Convert from Cartesian to spherical co-ordinates, following MATLAB's cart2sph.

    Parameters
    ----------
    x, y, z : floats
        Cartesian coordinates in the x, y and z dimensions.

    Returns
    -------
    azimuth : float
        The azimuth in degrees.
    elevation : float
        The elevation in degrees.
    r : float
        The radius.
    """
    azimuth = np.arctan2(y, x)
    elevation = np.arctan2(z, np.hypot(x, y))
    r = np.sqrt(x**2 + y**2 + z**2)
    return azimuth, elevation, r


def llh_to_s(
    longitude,
    latitude,
    height,
    central_latitude=90,
    central_longitude=0,
    radius_sphere=6371,
    scale_vertical=1,
):
    azimuth = np.deg2rad(longitude - central_longitude)
    elevation = np.deg2rad(latitude)
    radius = radius_sphere + scale_vertical * height / 1000
    sx, sy, sz = spherical_to_cartesian(azimuth, elevation, radius)
    stheta, srho = cartesian_to_polar(sx, sz)
    rot = (90 - central_latitude) * np.pi / 180
    stheta += rot
    sx, sz = polar_to_cartesian(stheta, srho)
    sx, sy = sy, -sx
    return sx, sy, sz


def s_to_llh(
    sx,
    sy,
    sz,
    central_latitude=90,
    central_longitude=0,
    radius_sphere=6371,
    scale_vertical=1,
):
    sx, sy = -sy, sx
    stheta, srho = cartesian_to_polar(sx, sz)
    rot_ = (90 - central_latitude) * np.pi / 180
    stheta -= rot_
    sx, sz = polar_to_cartesian(stheta, srho)
    azimuth_, elevation_, radius_ = cartesian_to_spherical(sx, sy, sz)
    longitude = np.rad2deg(azimuth_) + central_longitude
    latitude = np.rad2deg(elevation_)
    height = (radius_ - 6371) * 1000 / scale_vertical
    return longitude, latitude, height



"""
dreamcoat.ctd
=============
Read and process CTD data (specifically, from RV Pelagia).

Functions
---------
get_deep
    Create one-per-depth table from btl.
read_bl
    Read the bottle numbers and datetime from a bl file from the CTD.
read_btl
    Import a btl file and parse it into a usable format.
read_btl_dir
    Import all btl files in a directory and parse them into a usable format.
read_btl_raw
    Import a btl file from the CTD with minimal processing.
read_cnv
    Read the contents of a cnv file from a CTD cast.
read_ctd_bl_create_btl
    Create a new "btl file" from the 1 Hz data file with a selected averaging period.
get_downcast_upcast_1m_u_d
    Make logicals for downcast and upcast sections of a 1m-binned CTD file.
"""

import os
import textwrap
import pandas as pd
import numpy as np
from matplotlib import dates as mdates
from . import plot

ctd_renamer = {
    "scan": "scan",
    "prDM": "pressure",
    "t090C": "temperature",
    "CStarAt0": "beam_attenuation",
    "CStarTr0": "beam_transmission",
    "flC": "fluorescence_chelsea",
    "v2": "voltage_2",
    "ph": "pH_voltage",
    "sbeox0Mg/L": "oxygen_mgl",
    "sbeox0Mm/L": "oxygen_mml",
    "altM": "altimeter",
    "flECO-AFL": "fluorescence_wetlabs",
    "turbWETntu0": "turbidity",
    "timeS": "elapsed_seconds",
    "c0mS/cm": "conductivity",
    "sal00": "salinity",
    "potemp090C": "theta",
    "sigma-é00": "density",
    "depSM": "depth",
    "svCM": "sound_velocity",
    "bottle": "bottle",
    "c0S/m": "conductivity",
    "spar/sat/lin": "spar",
    "par/sat/log": "par",
    "sigma-t00": "sigma_theta",
}
ctd_Renamer = {k[0].upper() + k[1:]: v for k, v in ctd_renamer.items()}


def read_cnv(filename, station=None):
    """Read the contents of a cnv file from a CTD cast.

    Parameters
    ----------

    filename : str
        The filename (and path) to read.

    Returns
    -------
    pd.DataFrame
        The imported dataset.
    """
    # Find number of header lines to skip and get column headers and other info
    with open(filename, "r", encoding="unicode_escape", errors="replace") as f:
        skiprows = 0
        names = []
        ctd_Hz_line = ""
        while not ctd_Hz_line.startswith("*END*"):
            ctd_Hz_line = f.readline()
            # Get column headers
            if ctd_Hz_line.startswith("# name "):
                names.append(ctd_Hz_line.split(" = ")[1].split(":")[0])
            # Get latitude and longitude
            if ctd_Hz_line.startswith("* NMEA Latitude = "):
                latitude_txt = ctd_Hz_line.split(" = ")[1].split(" ")
                latitude = float(latitude_txt[0]) + float(latitude_txt[1]) / 60
                if latitude_txt[2] == "S\n":
                    latitude *= -1
            if ctd_Hz_line.startswith("* NMEA Longitude = "):
                longitude_txt = ctd_Hz_line.split(" = ")[1].split(" ")
                longitude = float(longitude_txt[0]) + float(longitude_txt[1]) / 60
                if longitude_txt[2] == "W\n":
                    longitude *= -1
            # Get datetime information
            if ctd_Hz_line.startswith("# start_time = "):
                start_time = pd.to_datetime(ctd_Hz_line.split(" = ")[1].split("[")[0])
            # Increment row counter
            skiprows += 1
    # Deal with duplicate names by appending "_2" to their second appearance
    names_duplicate = []
    for i, name in enumerate(names):
        if names.count(name) == 2:
            names[names.index(name, i + 1)] += "_2"
            names_duplicate.append(names[names.index(name + "_2", i + 1)])
    # Import the dataset with pandas
    ctd_Hz = (
        pd.read_table(
            filename,
            encoding="unicode_escape",
            encoding_errors="replace",
            names=names,
            skiprows=skiprows,
            sep=r"\s+",
        )
        .drop(columns=names_duplicate)
        .rename(columns=ctd_renamer)
    )
    # Add positional information to the dataframe
    if station is not None:
        ctd_Hz["station"] = int(station)
    ctd_Hz["datetime"] = start_time + np.arange(0, ctd_Hz.shape[0]).astype(
        "timedelta64[s]"
    )
    ctd_Hz["latitude"] = latitude
    ctd_Hz["longitude"] = longitude
    return ctd_Hz


def read_btl_raw(filename, colwidths=11, encoding="unicode_escape"):
    """Import a btl file from the CTD with minimal processing.

    Parameters
    ----------
    filename : str
        The btl filename (and path).
    colwidths : int, optional
        Width of each data column in the file, by default 11.
    encoding : str, optional
        Encoding of the .btl file, by default "unicode_escape".

    Returns
    -------
    pandas.DataFrame
        The btl file with minimal processing.
    dict
        Other potentially useful information.
    """
    btl_extras = {}
    # Find number of header lines to skip
    with open(filename, "r", encoding=encoding) as f:
        skiprows = -1
        btl_line = "*"
        while btl_line.startswith("*") or btl_line.startswith("#"):
            btl_line = f.readline()
            if btl_line.startswith("* NMEA Latitude"):
                btl_extras["latitude"] = btl_line.split(" = ")[1]
            elif btl_line.startswith("* NMEA Longitude"):
                btl_extras["longitude"] = btl_line.split(" = ")[1]
            skiprows += 1
    # This really awkward way of extracting the column headers is because sometimes they
    # run into each other (without whitespace in between) in the .btl files, which
    # prevents pandas from inferring the column widths correctly:
    headers = textwrap.wrap(
        btl_line.replace(" ", "_"), width=colwidths, break_on_hyphens=False
    )
    headers = [h.replace("_", "") for h in headers]
    headers = [ctd_Renamer[h] if h in ctd_Renamer else h for h in headers]
    headers.append("type")
    headers = {k: v for k, v in enumerate(headers)}
    # Now we can (hopefully) import the file
    btl_raw = pd.read_fwf(
        filename, skiprows=skiprows + 2, encoding=encoding, header=None
    ).rename(columns=headers)
    return btl_raw, btl_extras


def read_btl(filename, station=None, **kwargs):
    """Import a btl file from the CTD and format it nicely.

    Parameters
    ----------
    filename : str
        The btl filename (and path).
    station : str or int, optional
        The station number (which is added as a column), by default None.
    **kwargs
        Additional kwargs to pass to read_btl_raw().

    Returns
    -------
    pandas.DataFrame
        The nicely formatted bottle file.
    """
    # Read in the raw file
    btl, btl_extras = read_btl_raw(filename, **kwargs)
    # Parse datetime
    btl["date"] = btl.Date
    btl["time"] = btl.Date
    btl.loc[btl.bottle.isnull(), "date"] = np.nan
    btl.loc[~btl.bottle.isnull(), "time"] = np.nan
    btl["date"] = btl.date.ffill()
    btl["time"] = btl.time.bfill()
    btl["datetime"] = pd.to_datetime(btl.date + "T" + btl.time)
    btl["datenum"] = mdates.date2num(btl.datetime)
    btl.drop(columns=["Date", "date", "time"], inplace=True)
    # Put sdevs into the avg rows
    for c in ctd_renamer.values():  # first just copy the complete columns
        if c in btl and c != "bottle":
            btl[c + "_sd"] = btl[c]
    btl_avg = btl.type == "(avg)"
    btl_sdev = btl.type == "(sdev)"
    for c in ctd_renamer.values():
        if c in btl and c != "bottle":
            btl.loc[btl_sdev, c] = np.nan
            btl.loc[btl_avg, c + "_sd"] = np.nan
            btl[c + "_sd"] = btl[c + "_sd"].bfill()
    btl = btl[btl_avg].drop(columns="type").copy()
    # Parse longitude and latitude, if present in btl_extras
    if "latitude" in btl_extras:
        lat = btl_extras["latitude"].replace("\n", "")
        lat = lat.split()
        lat = " N".find(lat[2]) * float(lat[0]) + float(lat[1]) / 60
        btl_extras["latitude"] = lat
    if "longitude" in btl_extras:
        lon = btl_extras["longitude"].replace("\n", "")
        lon = lon.split()
        lon = " E".find(lon[2]) * float(lon[0]) + float(lon[1]) / 60
        btl_extras["longitude"] = lon
    # Add station and other extra columns
    btl["station"] = int(station)
    for k, v in btl_extras.items():
        btl[k] = v
    # Put data columns into alphabetical order
    btl_columns = btl.columns.sort_values()
    btl_starters = ["station", "bottle", *btl_extras.keys(), "datetime", "datenum"]
    btl_columns = [*btl_starters, *btl_columns.drop(btl_starters)]
    btl = btl[btl_columns]
    # Create station_bottle column and set as index
    btl["bottle"] = btl.bottle.astype(int)
    btl["station_bottle"] = btl.station.astype(str) + "-" + btl.bottle.astype(str)
    btl = btl.set_index("station_bottle")
    return btl


def read_btl_dir(btl_path, **kwargs):
    """Import all btl files in a directory into a single dataframe.

    Parameters
    ----------
    btl_path : str
        The path of the directory containing btl files.
    **kwargs
        Additional kwargs to pass to read_btl_raw().

    Returns
    -------
    pandas.DataFrame
        The nicely formatted bottle files.
    """
    # Get list of bottle files
    if not btl_path.endswith(os.sep):
        btl_path += os.sep
    btl_files = os.listdir(btl_path)
    btl_files = [f for f in btl_files if f.endswith(".btl")]
    # Get corresponding station for each file
    btl_files = {int(f.split("-")[1].split("CTD")[0]): f for f in btl_files}
    btl_stations = list(btl_files.keys())
    # Import the bottle files in ascending station order and concatenate into a single
    # dataframe
    btl_stations.sort()
    btl = []
    for station in btl_stations:
        btl.append(read_btl(btl_path + btl_files[station], station=station, **kwargs))
    btl = pd.concat(btl)
    return btl


def read_bl(filename_bl):
    """Read the bottle numbers and datetime from a bl file from the CTD.

    Parameters
    ----------
    filename_bl : str
        The filename (and path) for the bl file.

    Returns
    -------
    pd.DataFrame
        The bottle numbers and datetime of bottle firing.
    """
    with open(filename_bl, "r") as f:
        bl = f.read().splitlines()
    skiprows = 0
    while not bl[skiprows].startswith("1, 1,"):
        skiprows += 1
    bl = pd.read_csv(
        filename_bl,
        skiprows=skiprows,
        header=None,
        names=["bottle", "x", "datetime", "y", "z"],
    ).drop(columns=["x", "y", "z"])
    bl["datetime"] = pd.to_datetime(bl.datetime)
    bl["datenum"] = mdates.date2num(bl.datetime)
    return bl


def read_ctd_bl_create_btl(filename_cnv_1Hz, filename_bl, period="30s"):
    """Create a new "btl file" by averaging data in the 1 Hz data file over a certain
    period before each bottle was fired.

    Parameters
    ----------
    filename_cnv_1Hz : str
        The filename (and path) for the 1 Hz cnv file.
    filename_bl : str
        The filename (and path) for the bl file.

    Returns
    -------
    pd.DataFrame
        The new "btl file" with all variables averaged over the specified period before
        each bottle firing moment.
    """
    ctd = read_cnv(filename_cnv_1Hz)
    bl = read_bl(filename_bl)
    # Make new btl table
    btl = bl.set_index("bottle")
    # Add data from specified period before firing for each sensor
    time_before_firing = pd.Timedelta(period)
    for c in ctd.columns:
        if c not in btl:
            btl[c] = np.nan
            btl[c + "_std"] = np.nan
            btl[c + "_flag"] = np.nan
    ctd["bottle"] = 0
    for i, row in btl.iterrows():
        L = (ctd.datetime > row.datetime - time_before_firing) & (
            ctd.datetime <= row.datetime
        )
        ctd.loc[L, "bottle"] = i
        for c in ctd.columns:
            if c not in ["bottle", "datetime", "datenum"]:
                btl.loc[i, c] = ctd.loc[L, c].mean()
                btl.loc[i, c + "_std"] = ctd.loc[L, c].std()
                if np.isnan(btl.loc[i, c]):
                    btl.loc[i, c + "_flag"] = 9
                else:
                    btl.loc[i, c + "_flag"] = 2
    return ctd, btl


def get_deep(btl, cluster_bandwidth=1, cluster_vars=None, verbose=True):
    """Create one-per-depth table from btl.

    Parameters
    ----------
    btl : pd.DataFrame
        The bottle file, e.g. as imported with read_btl_dir.
    cluster_bandwidth : float, optional
        The cluster bandwidth in m, by default 1.
    cluster_vars : list, optional
        Which variables to compute clusters for, by default None, in which case, all
        possible columns in btl are used.
    verbose : bool, optional
        Whether to report on columns that cannot be clustered, by default True.

    Returns
    -------
    pd.DataFrame
        A table with one row per nominal bottle firing depth.
    """
    deep = {
        "station": [],
        "depth": [],
        "depth_std": [],
        "depth_count": [],
        "longitude": [],
        "latitude": [],
    }
    if cluster_vars is None:
        cluster_vars = list(btl.columns)
        if "bottle" in cluster_vars:
            cluster_vars.remove("bottle")
        cluster_vars = [v for v in cluster_vars if not v.endswith("_std")]
    for k in deep:
        if k in cluster_vars:
            cluster_vars.remove(k)
    for i, k in enumerate(cluster_vars):
        if np.isin(btl[k].dtype, [float, int]):
            deep[k] = []
            deep[k + "_std"] = []
            deep[k + "_count"] = []
        else:
            cluster_vars[i] = "__REMOVE__"
            if verbose:
                print('dc.ctd.get_deep(): Could not cluster column "{}".'.format(k))
    while "__REMOVE__" in cluster_vars:
        cluster_vars.remove("__REMOVE__")
    btl_stations = btl.station.unique()
    for s in btl_stations:
        L = btl.station == s
        cl = plot.get_clusters(
            btl[L].depth.to_numpy(),
            btl[L][["depth", *cluster_vars]].to_numpy(),
            cluster_bandwidth,
        )
        deep["station"].append(np.full(cl.x.size, s))
        for v in ["longitude", "latitude"]:
            deep[v].append(np.full(cl.x.size, btl[v][L].mean()))
        deep["depth"].append(cl.x)
        deep["depth_std"].append(cl.std_unbiased[:, 0])
        deep["depth_count"].append(cl.count[:, 0])
        for i, k in enumerate(cluster_vars):
            deep[k].append(cl.Y[:, i + 1])
            deep[k + "_std"].append(cl.std_unbiased[:, i + 1])
            deep[k + "_count"].append(cl.count[:, i + 1])
    deep = pd.DataFrame({k: np.concatenate(v) for k, v in deep.items()})
    deep["datetime"] = mdates.num2date(deep.datenum)
    deep["station_depth"] = (
        deep.station.astype(int).astype(str)
        + "-"
        + np.round(deep.depth).astype(int).astype(str)
    )
    deep = deep.set_index("station_depth")
    return deep


def get_downcast_upcast_1m_u_d(ctd_1m_u_d, inplace=True):
    """Make logicals for downcast and upcast sections of a 1m-binned CTD file.

    Parameters
    ----------
    ctd_1m_u_d : pd.DataFrame
        A 1m-binned CTD file including both downcast and upcast sections.
        May include data from multiple different stations.
        Must contain a filled "station" column, even if there is only one station.
    inplace : bool, optional
        Whether to add the new columns in-place, by default True

    Returns
    -------
    pd.DataFrame
        The ctd_1m_u_d dataframe with extra Boolean columns "downcast" and "upcast".
    """
    if not inplace:
        ctd_1m_u_d = ctd_1m_u_d.copy()
    ctd_1m_u_d["downcast"] = False
    ctd_1m_u_d["upcast"] = False
    for s, cast in ctd_1m_u_d.groupby("station"):
        sdiff = cast.depth.diff()
        sdiff.iloc[0] = 1
        sdiff_shift = sdiff.shift(-1)
        sdiff_shift.iloc[-1] = -1
        S = ctd_1m_u_d.station == s
        ctd_1m_u_d.loc[S, "downcast"] = sdiff > 0
        ctd_1m_u_d.loc[S, "upcast"] = sdiff_shift < 0
    if not inplace:
        return ctd_1m_u_d



import unittest
from dreamcoat.maps import vptree
import numpy as np


class TestVPTree(unittest.TestCase):

    def test_single_nearest_neighbor(self):
        dim = 10
        query = [0.5] * dim
        points, brute_force = brute_force_solution(20000, dim, query)
        tree = vptree.VPTree(points, euclidean)

        nearest = tree.get_nearest_neighbor(query)
        bf_nearest = brute_force[0]
        self.assertEqual(nearest[0], bf_nearest[0])
        self.assertTrue(all(n == b for n, b in zip(nearest[1], bf_nearest[1])))

    def test_nearest_neighbors(self):
        dim = 10
        query = [0.5] * dim
        points, brute_force = brute_force_solution(20000, dim, query)
        tree = vptree.VPTree(points, euclidean)

        for k in (1, 10, len(points)):
            tree_nearest = tree.get_n_nearest_neighbors(query, k)
            self.assertEqual(len(tree_nearest), k)
            brute_force_nearest = brute_force[:k]
            for nearest, bf_nearest in zip(tree_nearest, brute_force_nearest):
                self.assertEqual(nearest[0], bf_nearest[0])
                self.assertTrue(all(n == b for n, b in zip(nearest[1], bf_nearest[1])))

    def test_epsilon_search(self):
        dim = 10
        query = [0.5] * dim
        points, brute_force = brute_force_solution(20000, dim, query)
        tree = vptree.VPTree(points, euclidean)

        for eps in (-1, 0, 1, 2, 10):
            tree_nearest = sorted(tree.get_all_in_range(query, eps))
            brute_force_nearest = [point for point in brute_force if point[0] < eps]
            for nearest, bf_nearest in zip(tree_nearest, brute_force_nearest):
                self.assertEqual(nearest[0], bf_nearest[0])
                self.assertTrue(all(n == b for n, b in zip(nearest[1], bf_nearest[1])))

    def test_empty_points_raises_valueerror(self):
        self.assertRaises(ValueError, vptree.VPTree, [], euclidean)

    def test_zero_neighbors_raises_valueerror(self):
        tree = vptree.VPTree([1, 2, 3], euclidean)
        self.assertRaises(ValueError, tree.get_n_nearest_neighbors, [1], 0)


def euclidean(p1, p2):
    return np.sqrt(np.sum(np.power(p2 - p1, 2)))


def brute_force_solution(n, dim, query, dist=euclidean):
    points = np.random.randn(n, dim)
    brute_force = [(dist(query, point), point) for point in points]
    brute_force.sort()

    return points, brute_force


if __name__ == "__main__":
    unittest.main()



# This collection of classes just let me call something like compass.east to get 90 degrees.

from collections import namedtuple


class CompassSimple:
    """A class to use one of the 32 compass points inplace of degrees.  This was made so I could just call
    compass.east to get 90 degrees.
    """

    north = N = 0
    north_by_east = NbE = 11.25
    north_northeast = NNE = 22.5
    northeast_by_north = NEbN = 33.75
    northeast = NE = 45
    northeast_by_east = NEbE = 56.25
    east_northeast = ENE = 67.5
    east_by_north = EbN = 78.75
    east = E = 90
    east_by_south = EbS = 101.25
    east_southeast = ESE = 112.5
    southeast_by_east = SEbE = 123.75
    southeast = SE = 135
    southeast_by_south = SEbS = 146.25
    south_southeast = SSE = 157.5
    south_by_east = SbE = 168.75
    south = S = 180
    south_by_west = SbW = 191.25
    south_southwest = SSW = 202.5
    southwest_by_south = SWbS = 213.75
    southwest = SW = 225
    southwest_by_west = SWbW = 236.25
    west_southwest = WSW = 247.5
    west_by_south = WbS = 258.75
    west = W = 270
    west_by_north = WbN = 281.25
    west_northwest = WNW = 292.5
    northwest_by_west = NWbW = 303.75
    northwest = NW = 315
    northwest_by_north = NWbN = 326.25
    north_northwest = NNW = 337.5
    north_by_west = NbW = 348.75


class CompassComplex:
    """This class contains more information then the CompassSimple class.
    Recommended to use import CompassComplex as compass.
    """

    CompassPoint = namedtuple(
        "CompassPoint",
        ["Point", "Abbr", "WindPoint", "Minimum", "Direction", "Maximum"],
    )
    compass = dict()
    compass["north"] = compass["N"] = north = N = CompassPoint(
        "north", "N", "Tramontana", 354.375, 0.000, 5.625
    )
    compass["north by east"] = compass["NbE"] = north_by_east = NbE = CompassPoint(
        "north by east",
        "NbE",
        "Quarto di Tramontana verso Greco",
        5.625,
        11.250,
        16.875,
    )
    compass["north-northeast"] = compass["NNE"] = north_northeast = NNE = CompassPoint(
        "north-northeast", "NNE", "Greco-Tramontana", 16.875, 22.500, 28.125
    )
    compass["northeast by north"] = compass["NEbN"] = northeast_by_north = NEbN = (
        CompassPoint(
            "northeast by north",
            "NEbN",
            "Quarto di Greco verso Tramontana",
            28.125,
            33.750,
            39.375,
        )
    )
    compass["northeast"] = compass["NE"] = northeast = NE = CompassPoint(
        "northeast", "NE", "Greco", 39.375, 45.000, 50.625
    )
    compass["northeast by east"] = compass["NEbE"] = northeast_by_east = NEbE = (
        CompassPoint(
            "northeast by east",
            "NEbE",
            "Quarto di Greco verso Levante",
            50.625,
            56.250,
            61.875,
        )
    )
    compass["east-northeast"] = compass["ENE"] = east_northeast = ENE = CompassPoint(
        "east-northeast", "ENE", "Greco-Levante", 61.875, 67.500, 73.125
    )
    compass["east by north"] = compass["EbN"] = east_by_north = EbN = CompassPoint(
        "east by north", "EbN", "Quarto di Levante verso Greco", 73.125, 78.750, 84.375
    )
    compass["east"] = compass["E"] = east = E = CompassPoint(
        "east", "E", "Levante", 84.375, 90.000, 95.625
    )
    compass["east by south"] = compass["EbS"] = east_by_south = EbS = CompassPoint(
        "east by south",
        "EbS",
        "Quarto di Levante verso Scirocco",
        95.625,
        101.250,
        106.875,
    )
    compass["east-southeast"] = compass["ESE"] = east_southeast = ESE = CompassPoint(
        "east-southeast", "ESE", "Levante-Scirocco", 106.875, 112.500, 118.125
    )
    compass["southeast by east"] = compass["SEbE"] = southeast_by_east = SEbE = (
        CompassPoint(
            "southeast by east",
            "SEbE",
            "Quarto di Scirocco verso Levante",
            118.125,
            123.750,
            129.375,
        )
    )
    compass["southeast"] = compass["SE"] = southeast = SE = CompassPoint(
        "southeast", "SE", "Scirocco", 129.375, 135.000, 140.625
    )
    compass["southeast by south"] = compass["SEbS"] = southeast_by_south = SEbS = (
        CompassPoint(
            "southeast by south",
            "SEbS",
            "Quarto di Scirocco verso Ostro",
            140.625,
            146.250,
            151.875,
        )
    )
    compass["south-southeast"] = compass["SSE"] = south_southeast = SSE = CompassPoint(
        "south-southeast", "SSE", "Ostro-Scirocco", 151.875, 157.500, 163.125
    )
    compass["south by east"] = compass["SbE"] = south_by_east = SbE = CompassPoint(
        "south by east",
        "SbE",
        "Quarto di Ostro verso Scirocco",
        163.125,
        168.750,
        174.375,
    )
    compass["south"] = compass["S"] = south = S = CompassPoint(
        "south", "S", "Ostro", 174.375, 180.000, 185.625
    )
    compass["south by west"] = compass["SbW"] = south_by_west = SbW = CompassPoint(
        "south by west",
        "SbW",
        "Quarto di Ostro verso Libeccio",
        185.625,
        191.250,
        196.875,
    )
    compass["south-southwest"] = compass["SSW"] = south_southwest = SSW = CompassPoint(
        "south-southwest", "SSW", "Ostro-Libeccio", 196.875, 202.500, 208.125
    )
    compass["southwest by south"] = compass["SWbS"] = southwest_by_south = SWbS = (
        CompassPoint(
            "southwest by south",
            "SWbS",
            "Quarto di Libeccio verso Ostro",
            208.125,
            213.750,
            219.375,
        )
    )
    compass["southwest"] = compass["SW"] = southwest = SW = CompassPoint(
        "southwest", "SW", "Libeccio", 219.375, 225.000, 230.625
    )
    compass["southwest by west"] = compass["SWbW"] = southwest_by_west = SWbW = (
        CompassPoint(
            "southwest by west",
            "SWbW",
            "Quarto di Libeccio verso Ponente",
            230.625,
            236.250,
            241.875,
        )
    )
    compass["west-southwest"] = compass["WSW"] = west_southwest = WSW = CompassPoint(
        "west-southwest", "WSW", "Ponente-Libeccio", 241.875, 247.500, 253.125
    )
    compass["west by south"] = compass["WbS"] = west_by_south = WbS = CompassPoint(
        "west by south",
        "WbS",
        "Quarto di Ponente verso Libeccio",
        253.125,
        258.750,
        264.375,
    )
    compass["west"] = compass["W"] = west = W = CompassPoint(
        "west", "W", "Ponente", 264.375, 270.000, 275.625
    )
    compass["west by north"] = compass["WbN"] = west_by_north = WbN = CompassPoint(
        "west by north",
        "WbN",
        "Quarto di Ponente verso Maestro",
        275.625,
        281.250,
        286.875,
    )
    compass["west-northwest"] = compass["WNW"] = west_northwest = WNW = CompassPoint(
        "west-northwest", "WNW", "Maestro-Ponente", 286.875, 292.500, 298.125
    )
    compass["northwest by west"] = compass["NWbW"] = northwest_by_west = NWbW = (
        CompassPoint(
            "northwest by west",
            "NWbW",
            "Quarto di Maestro verso Ponente",
            298.125,
            303.750,
            309.375,
        )
    )
    compass["northwest"] = compass["NW"] = northwest = NW = CompassPoint(
        "northwest", "NW", "Maestro", 309.375, 315.000, 320.625
    )
    compass["northwest by north"] = compass["NWbN"] = northwest_by_north = NWbN = (
        CompassPoint(
            "northwest by north",
            "NWbN",
            "Quarto di Maestro verso Tramontana",
            320.625,
            326.250,
            331.875,
        )
    )
    compass["north-northwest"] = compass["NNW"] = north_northwest = NNW = CompassPoint(
        "north-northwest", "NNW", "Maestro-Tramontana", 331.875, 337.500, 343.125
    )
    compass["north by west"] = compass["NbW"] = north_by_west = NbW = CompassPoint(
        "north by west",
        "NbW",
        "Quarto di Tramontana verso Maestro",
        343.125,
        348.750,
        354.375,
    )

    full_name = [
        "north",
        "north by east",
        "north-northeast",
        "northeast by north",
        "northeast",
        "northeast by east",
        "east-northeast",
        "east by north",
        "east",
        "east by south",
        "east-southeast",
        "southeast by east",
        "southeast",
        "southeast by south",
        "south-southeast",
        "south by east",
        "south",
        "south by west",
        "south-southwest",
        "southwest by south",
        "southwest",
        "southwest by west",
        "west-southwest",
        "west by south",
        "west",
        "west by north",
        "west-northwest",
        "northwest by west",
        "northwest",
        "northwest by north",
        "north-northwest",
        "north by west",
    ]
    abbr = [
        "N",
        "NbE",
        "NNE",
        "NEbN",
        "NE",
        "NEbE",
        "ENE",
        "EbN",
        "E",
        "EbS",
        "ESE",
        "SEbE",
        "SE",
        "SEbS",
        "SSE",
        "SbE",
        "S",
        "SbW",
        "SSW",
        "SWbS",
        "SW",
        "SWbW",
        "WSW",
        "WbS",
        "W",
        "WbN",
        "WNW",
        "NWbW",
        "NW",
        "NWbN",
        "NNW",
        "NbW",
    ]

    def __init__(self):
        """
        Initializer for the Compass Class
        """
        pass

    def get_point(self, heading):
        return self._get_dir_name(heading, self.full_name)

    def get_abbr(self, heading):
        return self._get_dir_name(heading, self.abbr)

    def _get_dir_name(self, heading, type):
        """
        General case for finding the direction
        :param heading:
        :param type:
        :return:
        """
        heading = heading % 360
        for point in type:
            if self.compass[point].Direction == 0:
                if (
                    self.compass[point].Minimum <= heading <= 360.0
                    or 0.0 <= heading < self.compass[point].Maximum
                ):
                    return point
            elif self.compass[point].Minimum <= heading < self.compass[point].Maximum:
                return point


if __name__ == "__main__":
    import math

    print(CompassComplex.get_point(CompassComplex(), 200))
    print(CompassComplex.get_point(0))
    print(CompassComplex.get_point(360))
    print(CompassComplex.get_point(100))
    print(CompassComplex.get_point(2 * math.pi))
    print(CompassComplex.get_point(-15))
    print(CompassComplex.get_point(-360))
    print(CompassComplex.get_point(400))
    print(CompassComplex.get_point(-600))
    print(CompassComplex.N)
    print(CompassComplex.E.Direction)



import numpy as np
from dreamcoat.maps import vptree
import dreamcoat as dc


lat_range = (50, 55)
lon_range = (-5, 10)
coords = dc.maps.coastline_coords(lat_range, lon_range, resolution="110m")


def test_geodesic_distance():
    """Do we calculate the same distances as http://edwilliams.org/gccalc.htm ?"""
    dist0 = dc.maps.geodesic_distance((0, 1), (0, 0))
    assert isinstance(dist0, float)
    assert np.isclose(dist0, 110.57438855790893)
    dist1 = dc.maps.geodesic_distance((1, 0), (0, 0))
    assert isinstance(dist1, float)
    assert np.isclose(dist1, 111.31949077920639)
    dist2 = dc.maps.geodesic_distance((12, -23), (-135.8, 86.4))
    assert isinstance(dist2, float)
    assert np.isclose(dist2, 12885.15245722158)


def test_coastline_coords():
    """Do we return coordinates of the coastline(s) in the expected format?"""
    assert isinstance(coords, list)
    assert len(coords) > 0
    for coord in coords:
        assert isinstance(coord, tuple)
        assert len(coord) == 2
        assert isinstance(coord[0], float)
        assert isinstance(coord[1], float)


def test_build_vptree():
    """Can we build and use a vantage-point tree for the coastline?"""
    vpt = dc.maps.build_vptree(lat_range, lon_range, resolution="110m")
    assert isinstance(vpt, vptree.VPTree)
    nn0 = vpt.get_nearest_neighbor(coords[0])
    assert isinstance(nn0[0], float)
    assert nn0[0] == 0
    assert np.all(np.isclose(coords[0], nn0[1]))


# test_geodesic_distance()
# test_coastline_coords()
# test_build_vptree()



import pandas as pd
import numpy as np
import dreamcoat as dc
from cartopy import crs as ccrs

# Prepare adjacency edges for neutralocean surface calculation
edges = [
    [0, 24],
    [0, 27],
    [1, 8],
    [8, 19],
    [19, 24],
    [24, 27],
    [24, 35],
    [27, 35],
    [19, 40],
    [8, 46],
    [8, 50],
    [1, 59],
    [35, 40],
    [40, 46],
    [46, 50],
    [50, 59],
    [40, 70],
    [46, 63],
    [50, 74],
    [59, 112],
    [59, 117],
    [63, 70],
    [63, 74],
    [74, 112],
    [112, 117],
    [63, 93],
    [74, 89],
    [98, 117],
    [81, 86],
    [81, 112],
    [74, 81],
    [81, 89],
    [81, 98],
    [86, 89],
    [89, 93],
    [86, 98],
    [81, 117],
    [35, 70],
    [112, 126],
]
ctdz = pd.read_parquet("tests/data/ctdz.parquet")
graph = dc.CruiseGraph(ctdz, edges=edges)


def test_get_surfaces():
    station_ref = 1
    graph.get_surfaces(station_ref)
    assert station_ref in graph.surfaces_raw
    assert station_ref in graph.surfaces


def test_get_surfaces_all():
    graph.get_surfaces_all()
    for s in graph.stations.index:
        assert s in graph.surfaces


def test_plot_surface_map():
    extent = [-1, 7, 56.5, 64]
    crs = ccrs.EquidistantConic(central_longitude=np.mean([extent[0], extent[1]]))
    fig, ax = graph.plot_surface_map(1, 50, crs=crs, extent=extent)
    fig, ax = graph.plot_surface_map(8, 50, crs=crs, extent=extent)


# test_get_surfaces()
# test_get_surfaces_all()
# test_plot_surface_map()



import numpy as np


def _split_lon_ddm(lon_ddm):
    lon_split = lon_ddm.split("-")
    if len(lon_split) == 1:
        lon_split = [lon_ddm.split("°")[0], *lon_ddm.split("°")[1].split("'")]
    return lon_split


def _split_to_lon_dd(lon_split):
    return (float(lon_split[0]) + float(lon_split[1]) / 60) * dict(E=1, W=-1)[
        lon_split[2].upper()
    ]


def to_lon_dd(lon_ddm):
    return _split_to_lon_dd(_split_lon_ddm(lon_ddm))


def _split_lat_ddm(lat_ddm):
    lat_split = lat_ddm.split("-")
    if len(lat_split) == 1:
        lat_split = [lat_ddm.split("°")[0], *lat_ddm.split("°")[1].split("'")]
    return lat_split


def _split_to_lat_dd(lat_split):
    return (float(lat_split[0]) + float(lat_split[1]) / 60) * dict(N=1, S=-1)[
        lat_split[2].upper()
    ]


def to_lat_dd(lat_ddm):
    return _split_to_lat_dd(_split_lat_ddm(lat_ddm))


class LatLon:
    """A tool to convert between decimal degrees and degrees decimal minutes and to
    carry out basic calculations (e.g., add and subtract) using degrees decimal minutes.

    Parameters
    ----------
    latitude : float or str
        A latitude value.
          - If a float, should be in decimal degrees, with N positive.
          - If a str, should be in degrees decimal minutes, formatted either as (e.g.)
            "65°12.3'N" or "65-12.3-N".
    longitude : float or str
        A longitude value.
          - If a float, should be in decimal degrees, with E positive.
          - If a str, should be in degrees decimal minutes, formatted either as (e.g.)
            "65°12.3'E" or "65-12.3-E".

    Attributes
    ----------
    latitude_dd : float
        Latitude in decimal degrees, from -90 to 90 with N positive.
    latitude_r : int
        The floor of the latitude (degrees only).
    latitude_dm : float
        The decimal minutes of the latitude.
    latitude_dir : str
        The direction of the latitude ("N" or "S").
    longitude_dd : float
        Longitude in decimal degrees, from -180 to 180 with E positive.
    longitude_r : int
        The floor of the longitude (degrees only).
    longitude_dm : float
        The decimal minutes of the longitude.
    longitude_dir : str
        The direction of the longitude ("E" or "W").

    Methods
    -------
    to_dd : tuple of float
        Return (longitude, latitude) in decimal degrees.
    """

    def __init__(
        self,
        latitude=0,
        longitude=0,
    ):
        # Get values in decimal degrees
        if isinstance(longitude, str):
            self.longitude_dd = to_lon_dd(longitude.replace(" ", ""))
        elif isinstance(longitude, float) or isinstance(longitude, int):
            self.longitude_dd = longitude
        if isinstance(latitude, str):
            self.latitude_dd = to_lat_dd(latitude.replace(" ", ""))
        elif isinstance(latitude, float) or isinstance(latitude, int):
            self.latitude_dd = latitude
        # Adjust numbers to stay within valid ranges
        # (-90 to 90 for latitude, -180 to 180 for longitude)
        if self.latitude_dd > 90:
            self.latitude_dd = 180 - self.latitude_dd
            self.longitude_dd += 180
        self.keep_longitude_in_range()
        # Get rounded values and decimal minutes
        self.update_rounded_dm()

    def keep_longitude_in_range(self):
        if self.longitude_dd > 180:
            self.longitude_dd -= 360
        elif self.longitude_dd <= -180:
            self.longitude_dd += 360

    def update_rounded_dm(self):
        self.longitude_r = int(np.floor(np.abs(self.longitude_dd)))
        self.longitude_dm = 60 * (np.abs(self.longitude_dd) - self.longitude_r)
        self.longitude_dir = "EEW"[np.sign(self.longitude_dd).astype(int)]
        self.latitude_r = int(np.floor(np.abs(self.latitude_dd)))
        self.latitude_dm = 60 * (np.abs(self.latitude_dd) - self.latitude_r)
        self.latitude_dir = "NNS"[np.sign(self.latitude_dd).astype(int)]

    def to_dd(self, lon_first=True):
        """Return (longitude, latitude) in decimal degrees.

        Parameters
        ----------
        lon_first : bool, optional
            Whether to lead with longitude (True) or latitude (False), by default True.

        Returns
        -------
        tuple of float
            The (longitude, latitude) value in decimal degrees, with E and N positive
            (or (latitude, longitude) if not lon_first).
        """
        if lon_first:
            return self.longitude_dd, self.latitude_dd
        else:
            return self.latitude_dd, self.longitude_dd

    def to_ddm(self):
        """Return latitude and longitude in degrees decimal minutes.

        Returns
        -------
        str
            A nicely formatted string of the latitude and longitude values.
        """
        return "{:02.0f}°{:06.3f}'{}, {:03.0f}°{:06.3f}'{}".format(
            self.latitude_r,
            self.latitude_dm,
            self.latitude_dir,
            self.longitude_r,
            self.longitude_dm,
            self.longitude_dir,
        )

    def __repr__(self):
        return self.to_ddm() + " ({:.4f}, {:.4f})".format(
            self.longitude_dd, self.latitude_dd
        )

    def __add__(self, other):
        return LatLon(
            self.longitude_dd + other.longitude_dd, self.latitude_dd + other.latitude_dd
        )

    def __sub__(self, other):
        return LatLon(
            self.longitude_dd - other.longitude_dd, self.latitude_dd - other.latitude_dd
        )

    def __mul__(self, other):
        return LatLon(self.longitude_dd * other, self.latitude_dd * other)

    def __truediv__(self, other):
        return LatLon(self.longitude_dd / other, self.latitude_dd / other)


# def dd_to_ddm(dd):
#     """Convert decimal degrees into degrees decimal minutes.
#
#     Parameters
#     ----------
#     dd : float
#         (longitude, latitude) in decimal degrees.
#
#     Returns
#     -------
#     str
#         The position(s) in degrees decimal minutes.
#     """
#     single = dd.ndim == 1
#     if single:
#         dd = [dd]
#     ddm = []
#     for d in dd:
#         lon, lat = d
#         ddm.append(
#             "{:02.0f}°{:04.1f}'{}, {:03.0f}°{:04.1f}'{}".format(
#                 np.floor(np.abs(lat)),
#                 60 * (np.abs(lat) % 1),
#                 "SN"[int(lat > 0)],
#                 np.floor(np.abs(lon)),
#                 60 * (np.abs(lon) % 1),
#                 "WE"[int(lon > 0)],
#             )
#         )
#     if single:
#         ddm = ddm[0]
#     return ddm



import os

version_number = "0.4"
version_colour = "brown"
__version__ = "{}-{}".format(version_number, version_colour)


def hello():
    print(
        r"""
    ____    ____    U _____ u    _      __  __     ____   U  ___ u    _       _____   
   |  _"\U |  _"\ u \| ___"|/U  /"\  uU|' \/ '|uU /"___|   \/"_ \/U  /"\  u  |_ " _|  
  /| | | |\| |_) |/  |  _|"   \/ _ \/ \| |\/| |/\| | u     | | | | \/ _ \/     | |    
  U| |_| |\|  _ <    | |___   / ___ \  | |  | |  | |/__.-,_| |_| | / ___ \    /| |\   
   |____/ u|_| \_\   |_____| /_/   \_\ |_|  |_|   \____|\_)-\___/ /_/   \_\  u |_|U   
    |||_   //   \\_  <<   >>  \\    >><<,-,,-.   _// \\      \\    \\    >>  _// \\_  
   (__)_) (__)  (__)(__) (__)(__)  (__)(./  \.) (__)(__)    (__)  (__)  (__)(__) (__)
   """
    )
    print(
        "{:^88}".format("The joyful oceanographic seagoing expedition planning helper")
    )
    print(
        "{:^88}".format(
            "{} version ({})".format(version_colour.title(), version_number)
        )
    )


def get_dat_data(filename):
    """Find data from a file in the .dreamcoat folder, which should be in the user's
    home path, as returned by `os.path.expanduser('~')`.  If the file is not found, then
    the user is prompted to enter a value instead.

    Parameters
    ----------
    filename : str
        The name of the file, excluding the .dat extension.

    Returns
    -------
    str
        The value stored in the .dat file or provided by the user.
    """
    fn = os.sep.join((os.path.expanduser("~"), ".dreamcoat", "{}.dat".format(filename)))
    try:
        with open(fn, "r") as f:
            data = f.read().splitlines()[0]
    except FileNotFoundError:
        print("File {} not found!".format(fn))
        data = input("Please enter a value for {}: ".format(filename))
    return data



import math
from collections import namedtuple

pi = math.pi
# FAI Sphere
radius_earth_meters = 6371000
# Google Earth Sphere Radius
# radius_earth_meters = 6378137
eligible_units = ["meters", "kilometers", "miles", "feet", "yards", "nautical_miles"]
Units = namedtuple("Units", field_names=eligible_units)
radius_earth = Units(
    meters=radius_earth_meters,
    kilometers=radius_earth_meters / 1000,
    miles=radius_earth_meters / 1609.344,
    feet=radius_earth_meters * 3.28084,
    nautical_miles=radius_earth_meters / 1852,
    yards=radius_earth_meters * 1.09361,
)


#
# The following functions are redefined and passed to the math.* function because it is sometimes cumbersome to go
# math.* for all these functions.  Basically they are redefined here to make my life easier.
#


def sin(x):
    return math.sin(x)


def cos(x):
    return math.cos(x)


def asin(x):
    x = _test_domain(x)
    return math.asin(x)


def acos(x):
    x = _test_domain(x)
    return math.acos(x)


def atan2(y, x):
    return math.atan2(y, x)


def sqrt(x):
    return math.sqrt(x)


def _test_domain(x):
    if x > 1:
        return 1
    if x < -1:
        return -1
    return x



import dreamcoat as dc


def test_dd():
    lat = 3.3
    lon = -4.2
    ll = dc.LatLon(lat, lon)
    lon_out, lat_out = ll.to_dd()
    assert lon_out == lon
    assert lat_out == lat


def test_ddm():
    ll_dashes = dc.LatLon("67-12.55-N", "155-8.3-W")
    ll_fancy = dc.LatLon("67°12.55'N", "155°8.3'W")
    assert ll_dashes.latitude_dd == ll_fancy.latitude_dd
    assert ll_dashes.longitude_dd == ll_fancy.longitude_dd
    ll_str = ll_dashes.to_ddm()
    assert isinstance(ll_str, str)


def test_wrap():
    ll_180 = dc.LatLon(longitude=-90)
    ll_360 = dc.LatLon(longitude=270)
    assert ll_180.longitude_dd == ll_360.longitude_dd == -90


# test_dd()
# test_ddm()
# test_wrap()
